from abc import ABC, abstractmethod

import numpy as np
import torch as th

from mu_alpha_zero.General.memory import GeneralMemoryBuffer
from mu_alpha_zero.General.network import GeneralNetwork


class SearchTree(ABC):

    @abstractmethod
    def play_one_game(self, network, device: th.device,dir_path: str or None = None) -> tuple[list, int, int, int]:
        """
        Performs one game of the algorithm
        """
        pass

    @abstractmethod
    def search(self, network, state: np.ndarray, current_player: int or None, device: th.device,
               tau: float or None = None):
        """
        Performs MCTS search for given number of simulations.
        """

        pass

    @abstractmethod
    def make_fresh_instance(self):
        """
        Return new instance of this class.
        """
        pass

    @abstractmethod
    def step_root(self, action: int or None):
        """
        Steps the root node to the given action.
        """
        pass

    @abstractmethod
    def self_play(self, net: GeneralNetwork, device: th.device, num_games: int, memory: GeneralMemoryBuffer) -> tuple[
        int, int, int]:
        """
        Performs self play for given number of games.
        """
        pass



import io
import random
import sys

import numpy as np
import pygame as pg
import torch as th
from PIL import Image

from mu_alpha_zero.General.az_game import AlphaZeroGame


class TicTacToeGameManager(AlphaZeroGame):
    """
    This class is the game manager for the game of Tic Tac Toe and its variants.
    """

    def __init__(self, board_size: int, headless: bool, num_to_win=None) -> None:
        # TODO: Implement the possibility to play over internet using sockets.
        self.player = 1
        self.enemy_player = -1
        self.board_size = board_size
        self.board = self.initialize_board()
        self.headless = headless
        self.num_to_win = self.init_num_to_win(num_to_win)
        self.screen = self.create_screen(headless)

    def play(self, player: int, index: tuple) -> None:
        self.board[index] = player

    def init_num_to_win(self, num_to_win: int | None) -> int:
        if num_to_win is None:
            num_to_win = self.board_size
        if num_to_win > self.board_size:
            raise Exception("Num to win can't be greater than board size")
        return num_to_win

    def initialize_board(self):
        board = np.zeros((self.board_size, self.board_size), dtype=np.int8)
        return board

    def get_random_valid_action(self, observations: np.ndarray, **kwargs) -> list:
        valid_moves = self.get_valid_moves(observations)
        if len(valid_moves) == 0:
            raise Exception("No valid moves")
        return random.choice(valid_moves)

    def create_screen(self, headless):
        if headless:
            return

        pg.init()
        pg.display.set_caption("Tic Tac Toe")
        board_rect_size = self.board_size * 100
        screen = pg.display.set_mode((board_rect_size, board_rect_size))
        return screen

    def full_iterate_array(self, arr: np.ndarray, func: callable) -> list:
        """
        This function iterates over all rows, columns and the two main diagonals of supplied array,
        applies the supplied function to each of them and returns the results in a list.
        :param arr: a 2D numpy array.
        :param func: a callable function that takes a 1D array as input and returns a result.
        :return: A list of results.
        """
        results = []
        results.append("row")
        for row in arr:
            results.append(func(row.reshape(-1)))

        results.append("col")
        for col in arr.T:
            results.append(func(col.reshape(-1)))

        diags = [np.diag(arr, k=i) for i in range(-arr.shape[0] + 1, arr.shape[1])]
        flipped_diags = [np.diag(np.fliplr(arr), k=i) for i in range(-arr.shape[0] + 1, arr.shape[1])]
        diags.extend(flipped_diags)
        for idx, diag in enumerate(diags):
            if idx == 0:
                results.append("diag_left")
            elif idx == len(diags) // 2:
                results.append("diag_right")
            results.append(func(diag.reshape(-1)))

        # results.append("diag_left")
        # results.append(func(arr.diagonal().reshape(-1)))
        # results.append("diag_right")
        # results.append(func(np.fliplr(arr).diagonal().reshape(-1)))
        return results

    def eval_board(self, board: np.ndarray, check_end: bool = True) -> int | None:
        if self.is_board_full(board):
            return 0
        score = 0
        for i in range(self.num_to_win, self.num_to_win // 2, -1):
            current_won = self.check_partial_win(-1, self.num_to_win, board=board)
            opp_won = self.check_partial_win(1, self.num_to_win, board=board)
            if current_won:
                score += 1 * 2
            if opp_won:
                score -= 1
        if check_end:
            return None if self.game_result(-1, board) is None else score
        return score

    def full_iterate_array_all_diags(self, arr: np.ndarray, func: callable):

        results = []
        for row in arr:
            results.append(func(row.reshape(-1)))

        for col in arr.T:
            results.append(func(col.reshape(-1)))

        diags = [np.diag(arr, k=i) for i in range(-arr.shape[0] + 1, arr.shape[1])]
        flipped_diags = [np.diag(np.fliplr(arr), k=i) for i in range(-arr.shape[0] + 1, arr.shape[1])]
        diags.extend(flipped_diags)
        for diag in diags:
            # if diag.size < self.num_to_win:
            #     continue
            results.append(func(diag.reshape(-1)))

        return results

    def extract_all_vectors(self, board: np.ndarray):
        vectors = board.tolist() + board.T.tolist()
        vectors.extend([np.diag(board, k=i) for i in range(-board.shape[0] + 1, board.shape[1])])
        vectors.extend([np.diag(np.fliplr(board), k=i) for i in range(-board.shape[0] + 1, board.shape[1])])
        # pad vectors with -3's to have the same length as the board size and stack them
        vectors = [np.pad(vector, (0, self.board_size - len(vector)), constant_values=-3) for vector in vectors]
        return np.array(vectors)

    def check_win(self, player: int, board=None) -> bool:
        # if self.num_to_win == self.board_size:
        #     return self.check_full_win(player, board=board)
        # else:
        return self.check_partial_win(player, self.num_to_win, board=board)

    def check_full_win(self, player: int, board=None) -> bool:
        """
        This function checks if the supplied player has won the game with a full win (num_to_win == board_size).
        :param player: The player to check for (1 or -1).
        :param board: The board to check on. If None, the current board is used.
        :return: True if the player has won, False otherwise.
        """
        if board is None:
            board = self.get_board()
        matches = self.full_iterate_array(board, lambda part: np.all(part == player))
        for match in matches:
            if not isinstance(match, str) and match:
                return True

        return False

    def check_partial_win(self, player: int, n: int, board=None) -> bool:
        """
        This function checks if the supplied player has won the game with a partial win (num_to_win < board_size).

        :param player: The player to check for (1 or -1).
        :param n: The number of consecutive pieces needed to win.
        :param board: The board to check on. If None, the current board is used.
        :return: True if the player has won, False otherwise.
        """
        if board is None:
            board = self.get_board()
        matches = self.full_iterate_array_all_diags(board,
                                                    lambda part:
                                                    np.convolve((part == player), np.ones(n, dtype=np.int8),
                                                                "valid"))

        for match in matches:
            if np.any(match == n):
                return True

        return False

    def check_partial_win_vectorized(self, player: int, n: int, board=None) -> bool:
        if board is None:
            board = self.get_board()
        vectors = th.tensor(self.extract_all_vectors(board)).unsqueeze(0)
        weight = th.ones((1, 1, 1, n), dtype=th.long)
        vectors_where_player = th.where(vectors == player, 1, 0).long()
        res = th.nn.functional.conv2d(vectors_where_player, weight=weight)
        return th.any(res == n).item()

    def check_partial_win_to_index(self, player: int, n: int, board=None) -> dict[tuple, str] | dict:
        """
        This variation of check_partial_win returns the index of the first partial win found.
        The index being the index of the first piece in the winning sequence.
        :param player: The player to check for (1 or -1).
        :param n: The number of consecutive pieces needed to win.
        :param board: The board to check on. If None, the current board is used.
        :return: A dictionary containing the index and the position of the winning sequence.
        """
        if board is None:
            board = self.get_board()
        indices = self.full_iterate_array(board,
                                          lambda part:
                                          np.convolve((part == player), np.ones(n, dtype=int),
                                                      "valid"))
        indices = [x.tolist() if not isinstance(x, str) else x for x in indices]

        pos = "row"
        for vector in indices:
            if isinstance(vector, str):
                pos = vector
                continue
            for el in vector:
                if el == n:
                    vector_index = indices.index(vector)
                    pos_index = indices.index(pos)
                    # num_strings_before = len([x for index,x in enumerate(indices) if isinstance(x,str) and index < pos_index])
                    element_index = vector_index - (pos_index + 1) if vector_index > pos_index else indices.index(
                        vector, pos_index)
                    diag_idx = {
                        "index": ((self.board_size - 1) - element_index, 0)} if element_index < self.board_size else {
                        "index": (0, element_index - self.board_size)}
                    diag_idx["pos"] = pos
                    match pos:
                        case "row":
                            return {"index": (element_index, 0), "pos": pos}
                        case "col":
                            return {"index": (0, element_index), "pos": pos}
                        case "diag_left":
                            return diag_idx
                        case "diag_right":
                            return diag_idx

                        # case "diag_right":

        return {"index": None, "pos": None}

    def return_index_if_valid(self, index: tuple, return_on_fail: tuple = ()) -> tuple:
        if index[0] < 0 or index[0] >= self.board_size or index[1] < 0 or index[1] >= self.board_size:
            return return_on_fail
        return index

    def make_fresh_instance(self):
        return TicTacToeGameManager(self.board_size, self.headless, num_to_win=self.num_to_win)

    def get_previous(self, index: tuple, pos: str, n: int):
        if np.array(index).all() == 0:
            return index
        match pos:
            case "row":
                return self.return_index_if_valid((index[0], index[1] - n), return_on_fail=index)
            case "col":
                return self.return_index_if_valid((index[0] - n, index[1]), return_on_fail=index)
            case "diag_left":
                return self.return_index_if_valid((index[0] - n, index[1] - n), return_on_fail=index)
            case "diag_right":
                return self.return_index_if_valid((index[0] - n, index[1] + n), return_on_fail=index)

    def get_next(self, index: tuple, pos: str, n: int):
        if np.array(index).all() == self.board_size - 1:
            return index
        match pos:
            case "row":
                return self.return_index_if_valid((index[0], index[1] + n), return_on_fail=index)
            case "col":
                return self.return_index_if_valid((index[0] + n, index[1]), return_on_fail=index)
            case "diag_left":
                return self.return_index_if_valid((index[0] + n, index[1] + n), return_on_fail=index)
            case "diag_right":
                return self.return_index_if_valid((index[0] + n, index[1] - n), return_on_fail=index)

    def is_board_full(self, board=None) -> bool:
        if board is None:
            board = self.get_board()
        return np.all(board != 0)

    def get_board(self):
        return self.board.copy()

    def reset(self):
        self.board = self.initialize_board()
        return self.board.copy()

    def get_board_size(self):
        return self.board_size

    def render(self) -> bool:
        if self.headless:
            return False

        self.screen.fill((0, 0, 0))
        self._draw_board()
        for row in range(self.board_size):
            for col in range(self.board_size):
                if self.board[row][col] == self.player:
                    self._draw_circle(col * 100 + 50, row * 100 + 50)

                elif self.board[row][col] == self.enemy_player:
                    self._draw_cross(col * 100 + 50, row * 100 + 50)

        pg.event.pump()
        pg.display.flip()
        return True

    def _draw_circle(self, x, y) -> None:
        if self.headless:
            return
        pg.draw.circle(self.screen, "green", (x, y), 40, 1)

    def _draw_cross(self, x, y) -> None:
        if self.headless:
            return
        pg.draw.line(self.screen, "red", (x - 40, y - 40), (x + 40, y + 40), 1)
        pg.draw.line(self.screen, "red", (x + 40, y - 40), (x - 40, y + 40), 1)

    def _draw_board(self):
        for x in range(0, self.board_size * 100, 100):
            for y in range(0, self.board_size * 100, 100):
                pg.draw.rect(self.screen, (255, 255, 255), pg.Rect(x, y, 100, 100), 1)

    def is_empty(self, index: tuple) -> bool:
        return self.board[index] == 0

    def get_valid_moves(self, observation: np.ndarray, player: int or None = None) -> list:
        """
        Legal moves are the empty spaces on the board.
        :param observation: A 2D numpy array representing the current state of the game.
        :param player: The player to check for. Since the game is symmetric, this is ignored.
        :return: A list of legal moves.
        """
        legal_moves = []
        observation = observation.reshape(self.board_size, self.board_size)
        for row in range(self.board_size):
            for col in range(self.board_size):
                if observation[row][col] == 0:
                    legal_moves.append([row, col])
        return legal_moves

    def pygame_quit(self) -> bool:
        if self.headless:
            return False
        pg.quit()
        return True

    def get_click_coords(self):
        if self.headless:
            return
        mouse_pos = (x // 100 for x in pg.mouse.get_pos())
        if pg.mouse.get_pressed()[0]:  # left click
            return mouse_pos

    def get_human_input(self, board: np.ndarray):
        if self.headless:
            return
        while True:
            self.check_pg_events()
            if self.get_click_coords() is not None:
                x, y = self.get_click_coords()
                if board[y][x] == 0:
                    return y, x

            # time.sleep(1 / 60)

    def check_pg_events(self):
        if self.headless:
            return
        for event in pg.event.get():
            if event.type == pg.QUIT:
                self.pygame_quit()
                sys.exit(0)

    def network_to_board(self, move):
        """
        Converts an integer move from the network to a board index.
        :param move: An integer move selected from the network probabilities.
        :return: A tuple representing the board index (int,int).
        """
        return np.unravel_index(move, self.board.shape)

    @staticmethod
    def get_canonical_form(board, player) -> np.ndarray:
        return board * player

    def get_next_state(self, board: np.ndarray, action: int or tuple, player: int) -> np.ndarray:
        if isinstance(action, int):
            action = self.network_to_board(action)
        board_ = board.copy()
        board_[action] = player
        return board_

    def set_headless(self, val: bool):
        self.headless = val

    def get_invalid_actions(self, state: np.ndarray, player: int):
        mask = np.where(state == 0, 1, 0)
        return mask

    def __str__(self):
        return str(self.board).replace('1', 'X').replace('-1', 'O')
#
# if __name__ == "__main__":
#     sample_arr = np.array([[-1,1,-1,1,1],[-1,1,1,-1,-1],[1,-1,1,1,0],[-1,1,1,-1,0],[0,0,-1,1,1]])
#     game_manager = GameManager(5, True,num_to_win=3)
# res = game_manager.check_partial_win_vectorized(1,3,sample_arr)
# res = game_manager.game_result(-1,sample_arr)
# print(res)
# print(game_manager.check_partial_win(1,3,sample_arr))



import itertools
import random
import uuid
from collections import deque
from itertools import chain

import numpy as np
import numpy.random
import torch as th
from diskcache import Deque
from torch.utils.data import Dataset, DataLoader

from mu_alpha_zero.General.memory import GeneralMemoryBuffer
from mu_alpha_zero.General.utils import find_project_root
from mu_alpha_zero.Hooks.hook_manager import HookManager
from mu_alpha_zero.Hooks.hook_point import HookAt
from mu_alpha_zero.MuZero.lazy_arrays import LazyArray
from mu_alpha_zero.MuZero.utils import scale_action
from mu_alpha_zero.config import MuZeroConfig


class MemDataset(Dataset):
    def __init__(self, mem_buffer):
        self.mem_buffer = list(mem_buffer)

    def __len__(self):
        return len(self.mem_buffer)

    def __getitem__(self, idx):
        return self.mem_buffer[idx]


class MemBuffer(GeneralMemoryBuffer):
    def __init__(self, max_size, disk: bool = False, full_disk: bool = False, dir_path: str = None,
                 hook_manager: HookManager or None = None):
        self.max_size = max_size
        self.disk = disk
        self.full_disk = full_disk
        self.dir_path = dir_path
        self.hook_manager = hook_manager if hook_manager is not None else HookManager()
        self.buffer = self.init_buffer(dir_path)
        self.eval_buffer = self.init_buffer(dir_path)
        self.last_buffer_size = 0
        # self.priorities = []
        # self.eval_priorities = []
        self.is_disk = disk

    def add(self, experience, is_eval: bool = False):
        if not isinstance(experience, SingleGameData):
            raise ValueError("Experience must be a DataPoint instance")
        # if self.disk and not self.full_disk and not isinstance(experience.frame, LazyArray):
        #     frame = LazyArray(experience.frame, self.dir_path)
        #     experience.frame = frame

        if not is_eval:
            self.buffer.append(experience)
        else:
            self.eval_buffer.append(experience)

    def init_buffer(self, dir_path: str or None):
        if self.disk and self.full_disk:
            if dir_path is None:
                dir_path = f"{find_project_root()}/Pickles/Data"
            return Deque(maxlen=self.max_size, directory=dir_path)
        else:
            return deque(maxlen=self.max_size)

    def get_dir_path(self):
        return self.dir_path

    def train_length(self):
        return len(self.buffer)

    def add_list(self, experience_list, percent_eval: int = 10):
        if len(self.buffer) >= 10:
            prob = random.random()
            if prob < 0.1:
                # In 10% of cases, add to eval buffer
                for item in experience_list:
                    self.eval_buffer.append(item)
                return
        for item in experience_list:
            self.buffer.append(item)

    def sample(self, batch_size):
        return random.sample(self.buffer, batch_size)

    def shuffle(self, is_eval: bool = False):
        if is_eval:  # in-place shuffle
            random.shuffle(self.eval_buffer)
        else:
            random.shuffle(self.buffer)

    def batch(self, batch_size: int, is_eval: bool = False):
        buf = self.buffer if not is_eval else self.eval_buffer
        indeces = np.random.choice(np.arange(len(buf)), size=min(len(buf), batch_size),
                                   replace=False).flatten().tolist()
        items = [buf[i] for i in indeces]
        batch = [[items[i][x] for i in range(len(items))] for x in range(len(items[0]))]
        return [batch]

    def __call__(self, batch_size, is_eval: bool = False) -> list:
        return self.batch(batch_size, is_eval=is_eval)

    def __len__(self):
        return len(self.buffer)

    def eval_length(self):
        return len(self.eval_buffer)

    def batch_with_priorities(self, enable_per: bool, batch_size: int, config: MuZeroConfig, is_eval: bool = False):
        buf = self.buffer if not is_eval else self.eval_buffer
        # if recalculate_p_on_every_call:
        #     self.priorities = None
        # if self.priorities is None:
        #     self.priorities = self.calculate_priorities(enable_per, alpha, is_eval=is_eval)
        # return self.simple_sample(buf, batch_size, K)
        if enable_per:
            game_priorities = np.array([x.game_priority for x in buf])
        else:
            game_priorities = np.ones((len(buf),))

        game_priorities /= game_priorities.sum()
        sampled_indexes = np.random.choice(np.arange(len(buf)), replace=False, size=min(batch_size, len(buf)),
                                           p=game_priorities).tolist()
        positions = [buf[index].normalize(config).sample_position_with_unroll(config) for index in sampled_indexes]
        grad_scales = [x[1] for x in positions]
        positions = [x[0] for x in positions]
        if config.enable_per:
            weights = [positions[i].datapoints[0].priority * game_priorities[i] for i in range(len(positions))]
            weights = np.array(weights)
            weights = 1 / (weights * weights.shape[0]) ** config.beta
        else:
            weights = np.ones((len(positions),))
        return positions, grad_scales, game_priorities, weights.reshape(-1, 1)

    def reset_priorities(self):
        self.priorities = None

    def calculate_priorities(self, user_per: bool, alpha: float, is_eval: bool = False):
        buf = self.buffer if not is_eval else self.eval_buffer
        if not user_per:
            return np.ones((len(buf),)) / len(buf)
        ps = [abs(buf[i][1] - buf[i][2][2]) for i in range(len(buf))]
        ps = np.array(ps)
        ps = ps ** alpha
        ps = ps / ps.sum()
        self.hook_manager.process_hook_executes(self, self.calculate_priorities.__name__, __file__, HookAt.ALL,
                                                args=(ps,))
        return ps

    def simple_sample(self, buf: deque, batch_size: int, K: int) -> tuple:
        random_indexes = np.random.choice(np.arange(len(buf)),
                                          size=min(len(self.priorities) // K, max(batch_size // K, 1)),
                                          replace=False, p=self.priorities).tolist()
        batch = [list(itertools.islice(buf, i, i + K)) for i in random_indexes]
        pris = [self.priorities[i:i + K] for i in random_indexes]
        return list(chain.from_iterable(batch)), th.tensor(list(chain.from_iterable(pris)), dtype=th.float32)

    def make_fresh_instance(self):
        return MemBuffer(self.max_size, self.disk, self.full_disk, self.dir_path, hook_manager=self.hook_manager)

    def to_dataloader(self, batch_size):
        return DataLoader(MemDataset(self.buffer), batch_size=batch_size, shuffle=True, collate_fn=lambda x: x)

    def run_on_training_end(self):
        self.hook_manager.process_hook_executes(self, self.run_on_training_end.__name__, __file__, HookAt.ALL)

    def save(self):
        if not self.disk:
            buffer = list(self.buffer)
            th.save(buffer, f"{find_project_root()}/Pickles/memory_buffer.pt")
            return f"{find_project_root()}/Pickles/memory_buffer.pt"
        else:
            return self.buffer.directory

    def get_buffer(self):
        return self.buffer


class SingleGameData:
    def __init__(self):
        self.datapoints = []
        self.game_priority: float = float("-inf")

    def add_data_point(self, experience_point):
        self.datapoints.append(experience_point)

    def compute_initial_priorities(self, config: MuZeroConfig):
        for index, data_point in enumerate(self.datapoints):
            future_index = index + config.num_td_steps
            if future_index < len(self.datapoints):
                val = self.datapoints[future_index].v if self.datapoints[
                                                             future_index].player == data_point.player else - \
                    self.datapoints[future_index].v
                val = val * config.gamma ** config.num_td_steps
            else:
                val = 0
            for idx in range(index, future_index + 1):
                if idx >= len(self.datapoints):
                    break
                d_point = self.datapoints[idx]
                val += (
                           -d_point.reward if d_point.player == data_point.player else d_point.reward) * config.gamma ** idx
            if config.enable_per:
                self.datapoints[index].priority = abs(self.datapoints[index].v - val) ** config.alpha
                self.game_priority = max(self.game_priority, self.datapoints[index].priority)

            self.datapoints[index].v = val
        self.datapoints[0].v = 0

    def normalize(self, config: MuZeroConfig):
        if not config.enable_per:
            return self
        sum_ = 0
        for datapoint in self.datapoints:
            sum_ += datapoint.priority
        for datapoint in self.datapoints:
            datapoint.priority /= sum_
        return self

    def sample_position(self, config: MuZeroConfig):
        if config.enable_per:
            priorities = [x.priority for x in self.datapoints]
            index = numpy.random.choice(np.arange(len(self.datapoints)), size=(1,),
                                        p=priorities).tolist()[0]
        else:
            index = random.randint(0, len(self.datapoints) - 1)
        # index, max_datapoint = max([((i, x), x.priority) for i, x in enumerate(self.datapoints)], key=lambda x: x[1])[0]
        return self.datapoints[index], index

    def sample_position_with_unroll(self, config: MuZeroConfig):
        position, index = self.sample_position(config)
        game_data = SingleGameData()
        grad_scales = []
        for unroll_index in range(index, index + config.K + 1):
            grad_scales.append(min(len(self.datapoints) - index, config.K))
            if unroll_index < len(self.datapoints):
                game_data.add_data_point(self.datapoints[unroll_index])
            else:
                game_data.add_data_point(DataPoint(
                    np.divide(np.ones((config.net_action_size,)), config.net_action_size).tolist(),
                    0,
                    0,
                    random.randint(0, config.net_action_size - 1),
                    0,
                    None,
                    np.zeros((config.net_action_size,))
                ))
        return game_data, grad_scales


class DataPoint:
    def __init__(self, pi: dict, v: float, reward: float or None, move: int or None, player: int,
                 frame: np.ndarray or None,
                 action_mask: np.ndarray or None):
        self.pi = [x for x in pi.values()] if isinstance(pi, dict) else pi
        self.v = v
        self.reward = reward
        self.move = move
        self.player = player
        self.frame = frame
        self.priority = None
        self.action_mask = action_mask


class MuZeroFrameBuffer:
    def __init__(self, frame_buffer_size, noop_action: int, action_space_size: int, ignore_actions: bool = False):
        self.max_size = frame_buffer_size
        self.noop_action = noop_action
        self.ignore_actions = ignore_actions
        self.action_space_size = action_space_size
        self.buffers = {1: deque(maxlen=frame_buffer_size), -1: deque(maxlen=frame_buffer_size)}

    def add_frame(self, frame, action, player):
        if self.ignore_actions:
            self.buffers[player].append(frame)
            return
        self.buffers[player].append((frame, action))

    def concat_frames(self, player):
        if self.ignore_actions:
            return th.cat([th.tensor(frame, dtype=th.float32) for frame in self.buffers[player]], dim=2)
        frames_with_actions = [th.cat((th.tensor(frame, dtype=th.float32),
                                       th.full((frame.shape[0], frame.shape[1], 1), action, dtype=th.float32)), dim=2)
                               for frame, action in self.buffers[player]]
        # return th.tensor(np.array(frames_with_actions), dtype=th.float32)
        return th.cat(frames_with_actions, dim=2)

    def init_buffer(self, init_state, player):
        for _ in range(self.max_size):
            if not self.ignore_actions:
                self.buffers[player].append((init_state, scale_action(self.noop_action, self.action_space_size)))
            else:
                self.buffers[player].append(init_state)

    def __len__(self, player):
        return len(self.buffers[player])



import os

if "root.root" in os.listdir():
    exit(0)
    
    
os.makedirs("AlphaZeroTicTacToe",exist_ok=True)
os.chdir("AlphaZeroTicTacToe")
os.makedirs("Logs/ProgramLogs")
os.makedirs("Checkpoints/NetVersions/Temp")
os.makedirs("Checkpoints/Traces")


import json
import math
from typing import Callable

import numpy as np
import optuna
import torch as th
from PIL import Image

from mu_alpha_zero.config import MuZeroConfig


def add_actions_to_obs(observations: th.Tensor, actions: th.Tensor, dim=0):
    return th.cat((observations, actions), dim=dim)


def match_action_with_obs(observations: th.Tensor, action: int, config: MuZeroConfig):
    if config.is_atari:
        tensor_action = th.zeros((config.net_action_size,), dtype=th.float32, device=observations.device).scatter(0,
                                                                                                                  th.tensor(
                                                                                                                      action),
                                                                                                                  1)
        tensor_action = tensor_action.expand((observations.shape[1], observations.shape[2]))
    else:
        if config.actions_are == "columns":
            tensor_action = th.full((1, observations.size(1), observations.size(2)),
                                    scale_action(action, config.net_action_size),
                                    device=observations.device)

        elif config.actions_are == "rows":
            tensor_action = th.zeros((observations.shape[1],), device=observations.device).scatter(0, th.tensor(action,
                                                                                                                device=observations.device),
                                                                                                   1).unsqueeze(0)
            tensor_action = tensor_action.expand((1, observations.shape[1], observations.shape[2]))
        elif config.actions_are == "board":
            # unravel to 2d
            action = [action % observations.shape[1], action % observations.shape[2]]
            tensor_action = th.zeros((1, observations.shape[1], observations.shape[2]), device=observations.device)[
                action[0], action[1]] = 1
        else:
            raise ValueError("Invalid config.actions_are value.")
    return add_actions_to_obs(observations, tensor_action)


def match_action_with_obs_batch(observation_batch: th.Tensor, action_batch: list[int], config: MuZeroConfig):
    tensors = [match_action_with_obs(observation_batch[index], action_batch[index], config).unsqueeze(0) for index in
               range(len(action_batch))]
    return th.cat(tensors, dim=0)


def resize_obs(observations: np.ndarray, size: tuple[int, int], resize: bool) -> np.ndarray:
    if not resize:
        return observations
    obs = Image.fromarray(observations)
    obs = obs.resize(size)
    return np.array(obs)


def scale_state(state: np.ndarray, scale: bool):
    if not scale:
        return state
    # scales the given state to be between 0 and 1
    return state / 255


def scale_action(action: int, num_actions: int):
    # return action / (num_actions - 1)
    return (action + 1) / num_actions


def scale_hidden_state(hidden_state: th.Tensor):
    was_reshaped = False
    if len(hidden_state.shape) == 3:
        hidden_state = hidden_state.unsqueeze(0)
        was_reshaped = True
    max_ = hidden_state.view(hidden_state.size(0), hidden_state.size(1), -1).max(dim=2, keepdim=True)[0].unsqueeze(-1)
    min_ = hidden_state.view(hidden_state.size(0), hidden_state.size(1), -1).min(dim=2, keepdim=True)[0].unsqueeze(-1)
    max_min_dif = max_ - min_
    max_min_dif[max_min_dif < 1e-5] += 1e-5
    hidden_state = (hidden_state - min_) / max_min_dif
    return hidden_state.squeeze(0) if was_reshaped else hidden_state


def scale_reward_value(value: th.Tensor, e: float = 0.001):
    return th.sign(value) * (th.sqrt(th.abs(value) + 1) - 1) + e * value


def invert_scale_reward_value(value: th.Tensor, e: float = 0.001):
    return th.sign(value) * (
            ((th.sqrt(1 + 4 * 0.001 * (th.abs(value) + 1 + 0.001)) - 1) / (2 * 0.001))
            ** 2
            - 1
    )


def scale_reward(reward: float):
    return math.log(reward + 1, 5)


def muzero_loss(y_hat, y):
    return -th.sum(y * y_hat, dim=1).unsqueeze(1)


def scalar_to_support(x: th.Tensor, support_size: int, is_atari: bool):
    # x is of shape: [batch_size,1] at unroll step t
    if is_atari:
        x = scale_reward_value(x)
    x = th.clamp(x, -support_size, support_size)
    lower_p = 1 - (x - x.floor())
    upper_p = x - x.floor()
    support = th.zeros((x.size(0), 2 * support_size + 1), device=x.device)
    support.scatter_(1, (support_size + x.floor()).type(th.int64), lower_p)
    upper_index = (support_size + x.floor() + 1)
    upper_index[upper_index >= 2 * support_size] = 0
    support.scatter_(1, (
        upper_index).type(th.int64), upper_p)
    return support


def support_to_scalar(x: th.Tensor, support_size: int, is_atari: bool):
    support = th.arange(-support_size, support_size + 1, 1, dtype=x.dtype, device=x.device).unsqueeze(0)
    output = th.sum(x * support, dim=1)
    if is_atari:
        output = invert_scale_reward_value(output)
    return output.unsqueeze(1)


def mz_optuna_parameter_search(n_trials: int, storage: str or None, study_name: str, game,
                               muzero_config: MuZeroConfig, in_memory: bool = False, direction: str = "maximize",
                               arena_override=None, memory_override=None):
    def objective(trial: optuna.Trial):
        muzero_config.num_simulations = trial.suggest_int("num_mc_simulations", 60, 800)
        muzero_config.lr = trial.suggest_float("lr", 1e-5, 5e-2, log=True)
        muzero_config.tau = trial.suggest_float("temp", 0.5, 2)
        muzero_config.arena_tau = trial.suggest_float("arena_temp", 0, 2)
        muzero_config.c = trial.suggest_float("cpuct", 0.7, 2)
        muzero_config.l2 = trial.suggest_float("l2_norm", 1e-6, 6e-2)
        # muzero_config.frame_buffer_size = trial.suggest_int("frame_buffer_size", 10, 40)
        muzero_config.alpha = trial.suggest_float("alpha", 0.4, 2)
        muzero_config.beta = trial.suggest_float("beta", 0.1, 1)
        muzero_config.balance_term = trial.suggest_categorical("loss_scale", [1, 0.5])
        muzero_config.num_blocks = trial.suggest_int("num_blocks", 5, 32)
        muzero_config.K = trial.suggest_int("k", 2, 25)
        muzero_config.epochs = trial.suggest_int("epochs", 200, 500)

        muzero_config.self_play_games = 300
        muzero_config.num_iters = 1

        device = th.device("cuda" if th.cuda.is_available() else "cpu")
        trial.net_action_size = int(game.get_num_actions())
        network = MuZeroNet.make_from_config(muzero_config).to(device)
        tree = MuZeroSearchTree(game.make_fresh_instance(), muzero_config)
        net_player = NetPlayer(game.make_fresh_instance(), **{"network": network, "monte_carlo_tree_search": tree})
        if arena_override is None:
            arena = MzArena(game.make_fresh_instance(), muzero_config, device)
        else:
            arena = arena_override
        if memory_override is None:
            mem = MemBuffer(muzero_config.max_buffer_size, disk=True, full_disk=False,
                            dir_path=muzero_config.pickle_dir)
        else:
            mem = memory_override
        trainer = Trainer.create(muzero_config, game.make_fresh_instance(), network, tree, net_player, headless=True,
                                 arena_override=arena, checkpointer_verbose=False, memory_override=mem)
        trainer.train()
        mean = trainer.get_arena_win_frequencies_mean()
        trial.report(mean, muzero_config.num_iters)
        print(f"Trial {trial.number} finished with win freq {mean}.")
        del trainer
        del network
        del tree
        del net_player
        return mean

    from mu_alpha_zero.MuZero.MZ_Arena.arena import MzArena
    from mu_alpha_zero.MuZero.MZ_MCTS.mz_search_tree import MuZeroSearchTree
    from mu_alpha_zero.MuZero.Network.networks import MuZeroNet
    from mu_alpha_zero.trainer import Trainer
    from mu_alpha_zero.AlphaZero.Arena.players import NetPlayer
    from mu_alpha_zero.mem_buffer import MemBuffer

    muzero_config.show_tqdm = False
    if in_memory:
        study = optuna.create_study(study_name=study_name, direction=direction)
    else:
        if storage is None:
            raise ValueError("Storage can't be None if in_memory is False.")
        study = optuna.load_study(study_name=study_name, storage=storage)
    study.optimize(objective, n_trials=n_trials)
    with open(f"{muzero_config.checkpoint_dir}/study_params.json", "w") as file:
        json.dump(study.best_params, file)


def mask_invalid_actions(invalid_actions: np.ndarray, pi: np.ndarray):
    if np.sum(invalid_actions) == 0:
        print("No valid actions left.")
    pi = pi.reshape(-1) * invalid_actions.reshape(-1)
    return pi / pi.sum()


def mask_invalid_actions_batch(get_invalid_actions: Callable, pis: th.Tensor, players: list[int]):
    invalid_actions_ts = th.empty(pis.shape)
    for i, player in enumerate(players):
        invaid_actions = get_invalid_actions(pis[i], player)
        invalid_actions_ts[i] = invaid_actions
    return invalid_actions_ts


def get_opponent_action(game_state: np.ndarray, last_move: int, last_player: int, opponent: str, tree, net) -> tuple:
    if opponent == "muzero":
        res = tree.search(net, game_state, -last_player, th.device("cuda" if th.cuda.is_available() else "cpu"))
        return res[0], res[1][0]
    elif opponent == "minimax":
        return get_minimax_action_connect4(game_state[:,:,0], last_move, last_player)


def get_minimax_action_connect4(game_state: np.ndarray, last_move: int, last_player: int) -> tuple[dict, None]:
    from mu_alpha_zero.MuZero.MinimaxOpponent.board import Board
    from mu_alpha_zero.MuZero.MinimaxOpponent.player import PlayerMM
    board = Board.from_game_state(game_state, (last_player, last_move))
    depth = 5
    player = PlayerMM(depth, False)
    move = player.findMove(board)
    pi = [0.0] * game_state.shape[1]
    pi[move] = 1.0
    return {idx: el for idx, el in enumerate(pi)}, None



import atexit
import glob
import os
import time
from itertools import chain

import numpy as np
import torch as th
import torch.nn as nn
import torch.nn.functional as F
import wandb
from torch.nn.functional import mse_loss
from torch.nn.modules.module import T

from mu_alpha_zero.AlphaZero.checkpointer import CheckPointer
from mu_alpha_zero.AlphaZero.logger import Logger
from mu_alpha_zero.General.network import GeneralAlphZeroNetwork
from mu_alpha_zero.Hooks.hook_manager import HookManager
from mu_alpha_zero.Hooks.hook_point import HookAt
from mu_alpha_zero.config import AlphaZeroConfig, Config
from mu_alpha_zero.mem_buffer import MemBuffer
from mu_alpha_zero.shared_storage_manager import SharedStorage


class AlphaZeroNet(nn.Module, GeneralAlphZeroNetwork):
    def __init__(self, in_channels: int, num_channels: int, dropout: float, action_size: int, linear_input_size: int,
                 hook_manager: HookManager or None = None):
        super(AlphaZeroNet, self).__init__()
        self.in_channels = in_channels
        self.num_channels = num_channels
        self.dropout_p = dropout
        self.action_size = action_size
        self.linear_input_size = linear_input_size
        self.hook_manager = hook_manager if hook_manager is not None else HookManager()
        # Convolutional layers
        self.conv1 = nn.Conv2d(in_channels, num_channels, 3, padding=1)
        self.bn1 = nn.BatchNorm2d(num_channels)
        self.conv2 = nn.Conv2d(num_channels, num_channels, 3, padding=1)
        self.bn2 = nn.BatchNorm2d(num_channels)
        self.conv3 = nn.Conv2d(num_channels, num_channels, 3, padding=1)
        self.bn3 = nn.BatchNorm2d(num_channels)
        self.conv4 = nn.Conv2d(num_channels, num_channels, 3)
        self.bn4 = nn.BatchNorm2d(num_channels)

        # Fully connected layers
        # 4608 (5x5) or 512 (3x3) or 32768 (10x10) or 18432 (8x8) # or 8192 for atari (6x6)
        self.fc1 = nn.Linear(self.linear_input_size, 1024)
        self.fc1_bn = nn.BatchNorm1d(1024)
        self.fc2 = nn.Linear(1024, 512)
        self.fc2_bn = nn.BatchNorm1d(512)

        self.dropout = nn.Dropout(dropout)

        # Output layers
        self.pi = nn.Linear(512, action_size)  # probability head
        self.v = nn.Linear(512, 1)  # value head
        atexit.register(self.clear_traces)

    def forward(self, x, muzero: bool = True):
        if not muzero:
            x = x.unsqueeze(1)
        x = F.relu(self.bn1(self.conv1(x)))
        x = F.relu(self.bn2(self.conv2(x)))
        x = F.relu(self.bn3(self.conv3(x)))
        x = F.relu(self.bn4(self.conv4(x)))

        x = x.reshape(x.size(0), -1)  # Flatten

        x = F.relu(self.fc1_bn(self.fc1(x)))
        x = self.dropout(x)

        x = F.relu(self.fc2_bn(self.fc2(x)))
        x = self.dropout(x)

        pi = F.log_softmax(self.pi(x), dim=1)
        v = F.tanh(self.v(x))

        return pi, v

    @th.no_grad()
    def predict(self, x, muzero=True):
        pi, v = self.forward(x, muzero=muzero)
        pi = th.exp(pi)
        return pi.detach().cpu().numpy(), v.detach().cpu().numpy()

    def to_traced_script(self, board_size: int = 10):
        return th.jit.trace(self, th.rand(1, 256, board_size, board_size).cuda())

    def trace(self, board_size: int) -> str:
        traced = self.to_traced_script(board_size=board_size)
        path = "traced.pt"
        traced.save(path)
        return path

    def clear_traces(self) -> None:
        from mu_alpha_zero.General.utils import find_project_root
        for trace_file in glob.glob(f"{find_project_root()}/Checkpoints/Traces/*.pt"):
            os.remove(trace_file)

    def make_fresh_instance(self):
        return AlphaZeroNet(self.in_channels, self.num_channels, self.dropout_p, self.action_size,
                            self.linear_input_size)

    @staticmethod
    def make_from_config(config: AlphaZeroConfig, hook_manager: HookManager or None = None):
        return AlphaZeroNet(config.num_net_in_channels, config.num_net_channels, config.net_dropout,
                            config.net_action_size, config.az_net_linear_input_size, hook_manager=hook_manager)

    def train_net(self, memory_buffer: MemBuffer, alpha_zero_config: AlphaZeroConfig):
        from mu_alpha_zero.AlphaZero.utils import mask_invalid_actions_batch
        device = th.device("cuda" if th.cuda.is_available() else "cpu")
        losses = []
        optimizer = th.optim.Adam(self.parameters(), lr=alpha_zero_config.lr)
        memory_buffer.shuffle()
        for epoch in range(alpha_zero_config.epochs):
            for experience_batch in memory_buffer(alpha_zero_config.batch_size):
                if len(experience_batch) <= 1:
                    continue
                states, pi, v = zip(*experience_batch)
                states = th.tensor(np.array(states), dtype=th.float32, device=device)
                pi = th.tensor(np.array(pi), dtype=th.float32, device=device)
                v = th.tensor(v, dtype=th.float32, device=device).unsqueeze(1)
                pi_pred, v_pred = self.forward(states, muzero=False)
                masks = mask_invalid_actions_batch(states)
                loss = mse_loss(v_pred, v) + self.pi_loss(pi_pred, pi, masks, device)
                losses.append(loss.item())
                # self.summary_writer.add_scalar("Loss", loss.item(), i * epochs + epoch)

                optimizer.zero_grad()
                loss.backward()
                optimizer.step()
                self.hook_manager.process_hook_executes(self, self.train_net.__name__, __file__, HookAt.MIDDLE,
                                                        args=(experience_batch, loss.item(), epoch))

        self.hook_manager.process_hook_executes(self, self.train_net.__name__, __file__, HookAt.TAIL, args=(losses,))
        return sum(losses) / len(losses), losses

    def pi_loss(self, y_hat, y, masks, device: th.device):
        masks = masks.reshape(y_hat.shape).to(device)
        masked_y_hat = masks * y_hat
        return -th.sum(y * masked_y_hat) / y.size()[0]

    def to_shared_memory(self):
        for param in self.parameters():
            param.share_memory_()

    def run_at_training_end(self):
        self.hook_manager.process_hook_executes(self, self.run_at_training_end.__name__, __file__, HookAt.ALL)


class OriginalAlphaZeroNetwork(nn.Module, GeneralAlphZeroNetwork):

    def __init__(self, in_channels: int, num_channels: int, dropout: float, action_size: int,
                 linear_input_size: list[int], support_size: int,
                 state_linear_layers: int, pi_linear_layers: int, v_linear_layers: int, linear_head_hidden_size: int,
                 is_atari: bool,
                 latent_size: list[int] = [6, 6],
                 hook_manager: HookManager or None = None, num_blocks: int = 8, muzero: bool = False,
                 is_dynamics: bool = False, is_representation: bool = False):
        super(OriginalAlphaZeroNetwork, self).__init__()
        self.in_channels = in_channels
        self.num_channels = num_channels
        self.dropout_p = dropout
        self.action_size = action_size
        self.linear_input_size = linear_input_size
        self.support_size = support_size
        self.num_blocks = num_blocks
        self.muzero = muzero
        self.latent_size = latent_size
        self.is_dynamics = is_dynamics
        self.is_representation = is_representation
        self.state_linear_layers = state_linear_layers
        self.pi_linear_layers = pi_linear_layers
        self.v_linear_layers = v_linear_layers
        self.linear_head_hidden_size = linear_head_hidden_size
        self.is_atari = is_atari
        self.optimizer = None
        self.scheduler = None
        self.hook_manager = hook_manager if hook_manager is not None else HookManager()

        self.conv1 = nn.Conv2d(in_channels, num_channels, 3, padding=1)
        self.bn1 = nn.BatchNorm2d(num_channels)
        self.dropout = nn.Dropout(dropout)
        self.blocks = nn.ModuleList([OriginalAlphaZeroBlock(num_channels, num_channels) for _ in range(num_blocks)])
        if not is_representation:
            self.value_head = ValueHead(muzero, linear_input_size[0], support_size, num_channels, v_linear_layers,
                                        linear_head_hidden_size)
        else:
            self.value_head = th.nn.Identity()
        if is_dynamics or is_representation:
            self.policy_state_head = StateHead(linear_input_size[2], num_channels, latent_size, state_linear_layers,
                                               linear_head_hidden_size)
        else:
            self.policy_state_head = PolicyHead(action_size, linear_input_size[1], num_channels, pi_linear_layers,
                                                linear_head_hidden_size)

    def forward(self, x, muzero: bool = False, return_support: bool = False):
        if not muzero:
            x = x.unsqueeze(1)
        x = F.relu(self.bn1(self.conv1(x)))
        for block in self.blocks:
            x = block(x)
        # x = self.dropout(x)
        val_h_output = self.value_head(x)
        pol_h_output = self.policy_state_head(x)
        if self.is_representation:
            return pol_h_output
        if not return_support and self.muzero:
            from mu_alpha_zero.MuZero.utils import invert_scale_reward_value
            # multiply arange by softmax probabilities
            val_h_output = th.exp(val_h_output)
            support_range = th.arange(-self.support_size, self.support_size + 1, 1, dtype=th.float32,
                                      device=x.device).unsqueeze(0)
            output = th.sum(val_h_output * support_range, dim=1)
            if self.is_atari:
                output = invert_scale_reward_value(output)
            return pol_h_output, output.unsqueeze(1)
        return pol_h_output, val_h_output

    @th.no_grad()
    def predict(self, x, muzero=True):
        pi, v = self.forward(x, muzero=muzero)
        pi = th.exp(pi)
        return pi.detach().cpu().numpy(), v.detach().cpu().numpy()

    def pi_loss(self, y_hat, y, masks, device: th.device):
        masks = masks.reshape(y_hat.shape).to(device)
        masked_y_hat = masks * y_hat
        return -th.sum(y * masked_y_hat) / y.size()[0]

    def make_fresh_instance(self):
        return OriginalAlphaZeroNetwork(self.in_channels, self.num_channels, self.dropout_p, self.action_size,
                                        self.linear_input_size, self.support_size,
                                        self.state_linear_layers, self.pi_linear_layers, self.v_linear_layers,
                                        self.linear_head_hidden_size,
                                        self.is_atari,
                                        self.latent_size,
                                        hook_manager=self.hook_manager,
                                        num_blocks=self.num_blocks, muzero=self.muzero, is_dynamics=self.is_dynamics)

    @classmethod
    def make_from_config(cls, config: Config, hook_manager: HookManager or None = None):
        return OriginalAlphaZeroNetwork(config.num_net_in_channels, config.num_net_channels, config.net_dropout,
                                        config.net_action_size, config.az_net_linear_input_size,
                                        hook_manager=hook_manager,
                                        state_linear_layers=config.state_linear_layers,
                                        pi_linear_layers=config.pi_linear_layers,
                                        v_linear_layers=config.v_linear_layers,
                                        linear_head_hidden_size=config.linear_head_hidden_size,
                                        num_blocks=config.num_blocks, muzero=config.muzero,
                                        is_atari=config.is_atari,
                                        support_size=config.support_size, latent_size=config.net_latent_size)

    def train_net(self, memory_buffer, muzero_alphazero_config: Config) -> tuple[float, list[float]]:
        if memory_buffer.train_length() <= 1:
            return 0, []
        losses = []
        if self.optimizer is None:
            self.optimizer = th.optim.Adam(self.parameters(), lr=muzero_alphazero_config.lr,
                                           weight_decay=muzero_alphazero_config.l2)
        if muzero_alphazero_config.lr_scheduler is not None and self.scheduler is None:
            self.scheduler = muzero_alphazero_config.lr_scheduler(self.optimizer,
                                                                  **muzero_alphazero_config.lr_scheduler_kwargs)
        # memory_buffer.shuffle()
        for epoch in range(muzero_alphazero_config.epochs):
            for experience_batch in memory_buffer.batch(muzero_alphazero_config.batch_size):
                loss, v_loss, pi_loss = self.calculate_loss(experience_batch, muzero_alphazero_config)
                losses.append(loss.item())
                wandb.log({"combined_loss": loss.item(), "loss_v": v_loss.item(), "loss_pi": pi_loss.item()})
                self.optimizer.zero_grad()
                loss.backward()
                self.optimizer.step()
                if self.scheduler is not None:
                    self.scheduler.step()
                    wandb.log({"lr": self.scheduler.get_last_lr()[-1]})
                self.hook_manager.process_hook_executes(self, self.train_net.__name__, __file__, HookAt.MIDDLE,
                                                        args=(experience_batch, loss.item(), epoch))

        self.hook_manager.process_hook_executes(self, self.train_net.__name__, __file__, HookAt.TAIL, args=(losses,))
        return sum(losses) / len(losses), losses

    def eval_net(self, memory_buffer, muzero_alphazero_config: Config) -> None:
        if memory_buffer.eval_length() <= 1:
            return
        if self.optimizer is None:
            self.optimizer = th.optim.Adam(self.parameters(), lr=muzero_alphazero_config.lr,
                                           weight_decay=muzero_alphazero_config.l2)
        # memory_buffer.shuffle(is_eval=True)
        for epoch in range(muzero_alphazero_config.eval_epochs):
            for experience_batch in memory_buffer.batch(muzero_alphazero_config.batch_size, is_eval=True):
                loss, v_loss, pi_loss = self.calculate_loss(experience_batch, muzero_alphazero_config)
                wandb.log(
                    {"eval_combined_loss": loss.item(), "eval_loss_v": v_loss.item(), "eval_loss_pi": pi_loss.item()})

    def calculate_loss(self, experience_batch, muzero_alphazero_config):
        from mu_alpha_zero.AlphaZero.utils import mask_invalid_actions_batch
        device = th.device("cuda" if th.cuda.is_available() else "cpu")
        states, pi, v, _, masks = experience_batch[0], experience_batch[1], experience_batch[2], experience_batch[3], \
            experience_batch[4]
        pi = [[y for y in x.values()] if isinstance(x, dict) else x for x in pi]
        # game = [[y.frame,y.pi,y.v,y.action_mask] for y in experience_batch.datapoints]
        # states, pi, v, masks = zip(*game)
        states = th.tensor(np.array(states), dtype=th.float32, device=device)
        pi = th.tensor(np.array(pi), dtype=th.float32, device=device)
        v = th.tensor(v, dtype=th.float32, device=device).unsqueeze(1)
        masks = th.tensor(np.array(masks), dtype=th.float32, device=device)
        pi_pred, v_pred = self.forward(states, muzero=muzero_alphazero_config.muzero)
        v_loss = mse_loss(v_pred, v)
        pi_loss = self.pi_loss(pi_pred, pi, masks, device)
        loss = v_loss + pi_loss
        return loss, v_loss, pi_loss

    def continuous_weight_update(self, shared_storage: SharedStorage, alpha_zero_config: AlphaZeroConfig,
                                 checkpointer: CheckPointer or None,
                                 logger: Logger or None):
        wandb.init(project=alpha_zero_config.wandbd_project_name, name="Continuous Weight Update")
        alpha_zero_config.epochs = 1
        alpha_zero_config.eval_epochs = 50
        self.train()
        while shared_storage.train_length() < 200:
            time.sleep(5)
        params = shared_storage.get_stable_network_params()
        self.load_state_dict(params)
        for iter_ in range(alpha_zero_config.num_worker_iters):
            # print(iter_)
            # if not shared_storage.get_was_pitted():
            #     print("Waiting for pitting to finish")
            #     time.sleep(5)
            #     continue
            avg_iter_losses = self.train_net(shared_storage, alpha_zero_config)
            shared_storage.set_experimental_network_params(self.state_dict())
            shared_storage.set_training_iter(iter_)
            shared_storage.add_combined_loss(avg_iter_losses[0])
            # shared_storage.set_was_pitted(False)
            if iter_ % alpha_zero_config.eval_interval == 0 and iter_ != 0:
                self.eval_net(shared_storage, alpha_zero_config)
                checkpointer.save_checkpoint(self, self, self.optimizer, alpha_zero_config.lr, iter_, alpha_zero_config)
                print(f"Saved checkpoint at iteration {iter_}")


class OriginalAlphaZeroBlock(th.nn.Module):
    def __init__(self, in_channels: int, num_channels: int):
        super(OriginalAlphaZeroBlock, self).__init__()
        self.conv1 = nn.Conv2d(in_channels, num_channels, 3, padding=1)
        self.bn1 = nn.BatchNorm2d(num_channels)
        self.conv2 = nn.Conv2d(num_channels, num_channels, 3, padding=1)
        self.bn2 = nn.BatchNorm2d(num_channels)

    def forward(self, x):
        x_skip = x
        x = F.relu(self.bn1(self.conv1(x)))
        x = self.bn2(self.conv2(x))
        x += x_skip
        return F.relu(x)


class ValueHead(th.nn.Module):
    def __init__(self, muzero: bool, linear_input_size: int, support_size: int, num_channels: int, num_layers: int,
                 linear_hidden_size: int):
        super(ValueHead, self).__init__()
        self.conv = nn.Conv2d(num_channels, linear_hidden_size, 1)
        self.bn = nn.BatchNorm2d(linear_hidden_size)
        self.fc1 = nn.Linear(linear_input_size, 256)
        # self.fc = HeadLinear(256, 256, num_layers, linear_hidden_size)
        if muzero:
            self.fc2 = nn.Linear(256, 2 * support_size + 1)
            self.act = nn.LogSoftmax(dim=1)
        else:
            self.fc2 = nn.Linear(256, 1)
            self.act = nn.Tanh()

    def forward(self, x):
        x = F.relu(self.bn(self.conv(x)))
        x = x.reshape(x.size(0), -1)
        x = F.relu(self.fc1(x))
        # x = F.relu(self.fc(x))
        x = self.fc2(x)
        return self.act(x)


class PolicyHead(th.nn.Module):
    def __init__(self, action_size: int, linear_input_size_policy: int, num_channels: int, num_layers: int,
                 linear_hidden_size: int):
        super(PolicyHead, self).__init__()
        self.conv = nn.Conv2d(num_channels, linear_hidden_size, 1)
        self.bn = nn.BatchNorm2d(linear_hidden_size)
        self.fc1 = nn.Linear(linear_input_size_policy, 256)
        # self.fc = HeadLinear(256, 256, num_layers, linear_hidden_size)
        self.fc2 = nn.Linear(256, action_size)

    def forward(self, x):
        x = F.relu(self.bn(self.conv(x)))
        x = x.reshape(x.size(0), -1)
        x = F.relu(self.fc1(x))
        # x = F.relu(self.fc(x))
        x = self.fc2(x)
        return F.log_softmax(x, dim=1)


class StateHead(th.nn.Module):
    def __init__(self, linear_input_size: int, out_channels: int, latent_size: list[int], num_layers: int,
                 linear_hidden_size: int):
        super(StateHead, self).__init__()
        self.conv = nn.Conv2d(out_channels, out_channels, 1)
        self.bn = nn.BatchNorm2d(out_channels)
        # self.fc = HeadLinear(linear_input_size, out_channels, num_layers, linear_hidden_size)
        self.fc3 = nn.Linear(linear_input_size, latent_size[0] * latent_size[1] * out_channels)

    def forward(self, x):
        x = F.relu(self.bn(self.conv(x)))
        x = x.reshape(x.size(0), -1)
        # x = F.relu(self.fc(x))
        x = F.relu(self.fc3(x))
        return x


class HeadLinear(nn.Module):
    def __init__(self, in_channels: int, out_channels: int, num_layers: int, hidden_size: int):
        super(HeadLinear, self).__init__()
        self.fc1 = nn.Linear(in_channels, hidden_size)
        self.fc = nn.Sequential(*[nn.Linear(hidden_size, hidden_size) for _ in range(num_layers)])
        self.fc2 = nn.Linear(hidden_size, out_channels)

    def forward(self, x):
        x = F.relu(self.fc1(x))
        x = F.relu(self.fc(x))
        return self.fc2(x)



import copy
import gc
import random
import time

import wandb
from multiprocess import set_start_method

set_start_method("spawn", force=True)
from multiprocess.pool import Pool

import numpy as np
import torch as th
from mu_alpha_zero.General.memory import GeneralMemoryBuffer
from mu_alpha_zero.General.mz_game import MuZeroGame
from mu_alpha_zero.General.search_tree import SearchTree
from mu_alpha_zero.Hooks.hook_manager import HookManager
from mu_alpha_zero.Hooks.hook_point import HookAt
from mu_alpha_zero.MuZero.MZ_MCTS.mz_node import MzAlphaZeroNode
from mu_alpha_zero.MuZero.Network.networks import MuZeroNet
from mu_alpha_zero.MuZero.lazy_arrays import LazyArray
from mu_alpha_zero.MuZero.utils import match_action_with_obs, resize_obs, scale_action, scale_state, \
    scale_hidden_state, mask_invalid_actions
from mu_alpha_zero.config import MuZeroConfig
from mu_alpha_zero.mem_buffer import MuZeroFrameBuffer, SingleGameData, DataPoint
from mu_alpha_zero.shared_storage_manager import SharedStorage
from typing import Optional


class MuZeroSearchTree(SearchTree):

    def __init__(self, game_manager: MuZeroGame, muzero_config: MuZeroConfig,
                 hook_manager: Optional[HookManager] = None):
        self.game_manager = game_manager
        self.muzero_config = muzero_config
        self.hook_manager = hook_manager if hook_manager is not None else HookManager()
        self.buffer = self.init_frame_buffer()
        self.min_max_q = [float("inf"), -float("inf")]

    def init_frame_buffer(self):
        if self.muzero_config.enable_frame_buffer:
            return MuZeroFrameBuffer(self.muzero_config.frame_buffer_size, self.game_manager.get_noop(),
                                     self.muzero_config.net_action_size,
                                     ignore_actions=self.muzero_config.frame_buffer_ignores_actions)
        return MuZeroFrameBuffer(1, self.game_manager.get_noop(), self.muzero_config.net_action_size,
                                 ignore_actions=self.muzero_config.frame_buffer_ignores_actions)

    def play_one_game(self, network_wrapper: MuZeroNet, device: th.device, dir_path: Optional[str] = None,
                      calculate_avg_num_children: bool = False) -> list[SingleGameData]:

        self.buffer = self.init_frame_buffer()
        num_steps = self.muzero_config.num_steps
        frame_skip = self.muzero_config.frame_skip
        state = self.game_manager.reset()
        state = resize_obs(state, self.muzero_config.target_resolution, self.muzero_config.resize_images)
        state = scale_state(state, self.muzero_config.scale_state)
        player = 1
        self.buffer.init_buffer(state, player)
        if self.muzero_config.multiple_players:
            self.buffer.init_buffer(self.game_manager.get_state_for_passive_player(state, -player), -player)
        data = SingleGameData()
        frame = self.buffer.concat_frames(player).detach().cpu().numpy()
        data.add_data_point(
            DataPoint(None, None, 0, None, player, frame if dir_path is None else LazyArray(frame, dir_path),
                      self.game_manager.get_invalid_actions(player)))
        game_length = 0
        for step in range(num_steps):
            game_length += 1
            pi, (v, latent) = self.search(network_wrapper, state, player, device, calculate_avg_num_children=(
                    calculate_avg_num_children and step == 0))
            move = self.game_manager.select_move(pi, tau=self.muzero_config.tau)
            # _, pred_v = network_wrapper.prediction_forward(latent.unsqueeze(0), predict=True)
            state, rew, done = self.game_manager.frame_skip_step(move, player, frame_skip=frame_skip)
            state = resize_obs(state, self.muzero_config.target_resolution, self.muzero_config.resize_images)
            state = scale_state(state, self.muzero_config.scale_state)

            # data.append(
            #     (pi, v, (rew, move, float(pred_v[0]), player),
            #      ))
            if self.muzero_config.multiple_players:
                player = -player
            self.buffer.add_frame(state, scale_action(move, self.game_manager.get_num_actions()), player)
            self.buffer.add_frame(self.game_manager.get_state_for_passive_player(state, -player),
                                  scale_action(move, self.game_manager.get_num_actions()), -player)
            data.datapoints[-1].v = v
            data.datapoints[-1].pi = [x for x in pi.values()]
            data.datapoints[-1].move = move
            frame = self.buffer.concat_frames(player).detach().cpu().numpy()
            data.add_data_point(
                DataPoint(np.ones((len(data.datapoints[-1].pi))) / len(data.datapoints[-1].pi), 0, rew,
                          random.randint(0, len(data.datapoints[-1].pi) - 1), player,
                          frame if dir_path is None else LazyArray(frame, dir_path),
                          self.game_manager.get_invalid_actions(player)))
            if done:
                # time.sleep(1)
                # print(player)
                # print(v)
                break
            # self.game_manager.render()
            # time.sleep(0.5)

        try:
            wandb.log({"Game length": game_length})
        except Exception:
            pass
        data.compute_initial_priorities(self.muzero_config)
        return [data]

    def search(self, network_wrapper, state: np.ndarray, current_player: Optional[int], device: th.device,
               tau: Optional[float] = None, calculate_avg_num_children: bool = False, use_state_directly: bool = False):
        self.min_max_q = [float("inf"), -float("inf")]
        if self.buffer.__len__(current_player) == 0 and not use_state_directly:
            self.buffer.init_buffer(state, current_player)
        num_simulations = self.muzero_config.num_simulations
        if tau is None:
            tau = self.muzero_config.tau

        games = [copy.deepcopy(self.game_manager)] if self.muzero_config.use_true_game_state_in_tree else []
        root_node = MzAlphaZeroNode(current_player=current_player)
        # print(self.buffer.buffers[current_player][-1][0][:,:,0])
        game_state = state if use_state_directly else self.buffer.concat_frames(current_player)
        state_ = network_wrapper.representation_forward(
            game_state.permute(2, 0, 1).unsqueeze(0).to(device)).squeeze(0)
        if self.muzero_config.scale_hidden_state:
            state_ = scale_hidden_state(state_)
        pi, v = network_wrapper.prediction_forward(state_.unsqueeze(0), predict=True)
        if self.muzero_config.dirichlet_alpha > 0:
            pi = pi + np.random.dirichlet([self.muzero_config.dirichlet_alpha] * self.muzero_config.net_action_size)
        pi = mask_invalid_actions(self.game_manager.get_invalid_actions(current_player), pi)
        pi = pi.flatten().tolist()
        root_node.expand_node(state_, pi, 0)
        for simulation in range(num_simulations):
            current_node = root_node
            path = [current_node]
            action = None
            while current_node.was_visited():
                done = False
                current_node, action = current_node.get_best_child(self.min_max_q[0], self.min_max_q[1],
                                                                   self.muzero_config.gamma,
                                                                   self.muzero_config.multiple_players,
                                                                   c=self.muzero_config.c, c2=self.muzero_config.c2)
                path.append(current_node)
                if self.muzero_config.use_true_game_state_in_tree:
                    game = copy.deepcopy(games[-1])
                    _, rew, done = game.get_next_state(action, current_node.parent().current_player)
                    games.append(game)

            # action = scale_action(action, self.game_manager.get_num_actions())
            if not done:
                current_node_state_with_action = match_action_with_obs(current_node.parent().state, action,
                                                                       self.muzero_config)
                next_state, reward = network_wrapper.dynamics_forward(current_node_state_with_action.unsqueeze(0),
                                                                      predict=True)
                if self.muzero_config.scale_hidden_state:
                    next_state = scale_hidden_state(next_state)
                reward = reward[0][0]
                pi, v = network_wrapper.prediction_forward(next_state.unsqueeze(0), predict=True)
                if self.muzero_config.use_true_game_state_in_tree:
                    reward = rew
                    pi = mask_invalid_actions(games[-1].get_invalid_actions(current_node.current_player), pi)
                pi = pi.flatten().tolist()
                v = v.flatten().tolist()[0]
                current_node.expand_node(next_state, pi, reward)
            else:
                if self.muzero_config.multiple_players:
                    v = -1
                else:
                    v = rew
            self.backprop(v, path, games)

        action_probs = root_node.get_self_action_probabilities()
        root_val_latent = (root_node.get_self_value(), root_node.get_latent())
        self.hook_manager.process_hook_executes(self, self.search.__name__, __file__, HookAt.TAIL,
                                                args=(action_probs, root_val_latent, root_node))
        # if calculate_avg_num_children:
        #     num_nodes = self.get_num_nodes(root_node)
        #     wandb.log({"Number of non-leaf nodes": num_nodes})
        root_node = None
        return action_probs, root_val_latent

    def backprop(self, v: float, path: list, games: list):
        # G = v
        G_node = v
        gamma = self.muzero_config.gamma
        for node in reversed(path):
            node.times_visited += 1
            if self.muzero_config.multiple_players:
                # G = node.reward + gamma * (
                #     -G)  # G should be from the perspective of the parent as the parent is selecting from the children based on what's good for them.
                node.total_value += G_node
                self.update_min_max_q(node.reward - node.get_self_value())
                G_node = node.reward + gamma * (-G_node)
            else:
                node.total_value += G_node
                self.update_min_max_q(node.reward + node.get_self_value())
                G_node = node.reward + gamma * G_node
            if self.muzero_config.use_true_game_state_in_tree and node.select_probability != 0:
                games.pop()
            # node.update_q(G_node)

    def self_play(self, net: MuZeroNet, device: th.device, num_games: int, memory: GeneralMemoryBuffer) -> tuple[
        int, int, int]:
        for game in range(num_games):
            game_results = self.play_one_game(net, device, calculate_avg_num_children=game == num_games - 1)
            memory.add_list(game_results)

        return None, None, None

    def make_fresh_instance(self):
        return MuZeroSearchTree(self.game_manager.make_fresh_instance(), copy.deepcopy(self.muzero_config),
                                hook_manager=copy.deepcopy(self.hook_manager))

    def step_root(self, action: int or None):
        # I am never reusing the tree in MuZero.
        pass

    def update_min_max_q(self, q):
        self.min_max_q[0] = min(self.min_max_q[0], q)
        self.min_max_q[1] = max(self.min_max_q[1], q)

    def get_num_nodes(self, root_node: MzAlphaZeroNode):
        num_nodes = 0
        if len(root_node.children) > 0:
            num_nodes = 1
        for child in root_node.children.values():
            nn = self.get_num_nodes(child)
            num_nodes += nn

        return num_nodes

    @staticmethod
    def parallel_self_play(nets: list, trees: list, memory: GeneralMemoryBuffer, device: th.device, num_games: int,
                           num_jobs: int):
        with Pool(num_jobs) as p:
            if memory.is_disk and memory.full_disk:
                results = p.starmap(p_self_play, [
                    (nets[i], trees[i], copy.deepcopy(device), num_games // num_jobs, copy.deepcopy(memory), None) for i
                    in range(len(nets))])
            else:
                results = p.starmap(p_self_play, [
                    (nets[i], trees[i], copy.deepcopy(device), num_games // num_jobs, None, memory.dir_path) for i in
                    range(len(nets))])
        for result in results:
            memory.add_list(result)

        return None, None, None

    @staticmethod
    def start_continuous_self_play(nets: list, trees: list, shared_storage: SharedStorage,
                                   device: th.device, config: MuZeroConfig, num_jobs: int, num_worker_iters: int):
        pool = Pool(num_jobs)
        for i in range(num_jobs):
            pool.apply_async(c_p_self_play, args=(
                nets[i], trees[i], copy.deepcopy(device), config, i, shared_storage, num_worker_iters,
                shared_storage.get_dir_path())
                             )

        return pool

    @staticmethod
    def reanalyze(net, tree, device, shared_storage: SharedStorage, config: MuZeroConfig):
        wandb.init(project=config.wandbd_project_name, name="Reanalyze")
        net = net.to(device)
        while len(shared_storage.get_buffer()) < 10_000:
            time.sleep(5)
        for iter_ in range(config.num_worker_iters):
            data = random.choice(shared_storage.get_buffer())
            net.eval()
            if shared_storage.get_experimental_network_params() is not None:
                net.load_state_dict(shared_storage.get_experimental_network_params())
            else:
                net.load_state_dict(shared_storage.get_stable_network_params())
            wandb.log({"reanalyze_iteration": iter_})
            tree = tree.make_fresh_instance()
            for data_point in data.datapoints:
                if isinstance(data_point.frame, LazyArray):
                    frame = data_point.frame.load_array()
                else:
                    frame = data_point.frame

                state = th.tensor(frame, device=device, dtype=th.float32)
                pi, (v, _) = tree.search(net, state, data_point.player, device, use_state_directly=True)
                data_point.v = v
                data_point.pi = [x for x in pi.values()]
            data.compute_initial_priorities(config)

    def run_on_training_end(self):
        self.hook_manager.process_hook_executes(self, self.run_on_training_end.__name__, __file__, HookAt.ALL)


def p_self_play(net, tree, dev, num_g, mem, dir_path: Optional[str] = None):
    data = []
    for game in range(num_g):
        game_results = tree.play_one_game(net, dev, dir_path=dir_path, calculate_avg_num_children=game == num_g - 1)
        if mem is not None:
            mem.add_list(game_results)
        else:
            data.extend(game_results)
    return data


def c_p_self_play(net, tree, device, config: MuZeroConfig, p_num: int, shared_storage: SharedStorage,
                  num_worker_iters: int,
                  dir_path: Optional[str] = None):
    if p_num == 0:
        wandb.init(project=config.wandbd_project_name, name="Self play")
    net = net.to(device)
    for iter_ in range(num_worker_iters):
        # for game in range(num_g):
        # if not shared_storage.get_was_pitted():
        #     # If the network was not yet decided on, slow down the process so the data won't get overpopulated with current params.
        #     time.sleep(5)
        if shared_storage.get_experimental_network_params() is None:
            params = shared_storage.get_stable_network_params()
        else:
            params = shared_storage.get_experimental_network_params()
        net.eval()
        net.load_state_dict(params)
        game_results = tree.play_one_game(net, device, dir_path=dir_path)
        shared_storage.add_list(game_results)



import enum


class HookAt(enum.Enum):
    HEAD = "head"
    TAIL = "tail"
    MIDDLE = "middle"
    ALL = "all"


class HookPoint:
    def __init__(self, at: HookAt, file: str, fn_name: str):
        self.__at = at
        self.__file = file
        self.__function_name = fn_name

    @property
    def at(self) -> HookAt:
        return self.__at

    @property
    def file(self) -> str:
        return self.__file

    @property
    def function_name(self) -> str:
        return self.__function_name

    def here(self, file: str, function_name: str, at: HookAt):
        return file == self.file and function_name == self.function_name and (at == self.at or at == HookAt.ALL)



class Board:
    ''' class to store and manipulate connect 4 game states '''

    def __init__(self, width=7, height=6):
        self.w = width
        self.h = height
        self.board_state = [0, 0]
        self.col_heights = [(height + 1) * i for i in range(width)]
        self.moves = 0
        self.history = []
        self.node_count = 0
        # self.buffer = self.__get_buffer()
        self.bit_shifts = self.__get_bit_shifts()
        self.base_search_order = self.__get_base_search_order()

    def __repr__(self):
        state = []
        for i in range(self.h):                         # row
            row_str = ''
            for j in range(self.w):                     # col
                pos = 1 << (self.h + 1) * j + i
                if self.board_state[0] & pos == pos:
                    row_str += 'x '
                elif self.board_state[1] & pos == pos:
                    row_str += 'o '
                else:
                    row_str += '. '
            state.append(row_str)
        state.reverse()         # inverted orientation more readable
        return '\n'.join(state)

    def get_current_player(self):
        ''' returns current player: 0 or 1 (0 always plays first) '''
        return self.moves & 1

    def get_opponent(self):
        ''' returns opponent to current player: 0 or 1 '''
        return (self.moves + 1) & 1

    def get_search_order(self):
        ''' returns column search order containing playable columns only '''
        col_order = filter(self.can_play, self.base_search_order)
        return sorted(col_order, key=self.__col_sort, reverse=True)

    def get_mask(self):
        ''' returns bitstring of all occupied positions '''
        return self.board_state[0] | self.board_state[1]

    def get_key(self):
        ''' returns unique game state identifier '''
        return self.get_mask() + self.board_state[self.get_current_player()]

    def can_play(self, col):
        ''' returns true if col (zero indexed) is playable '''
        return not self.get_mask() & 1 << (self.h + 1) * col + (self.h - 1)

    def play(self, col):
        player = self.get_current_player()
        move = 1 << self.col_heights[col]
        self.col_heights[col] += 1
        self.board_state[player] |= move
        self.history.append(col)
        self.moves += 1

    def backtrack(self):
        opp = self.get_opponent()
        col = self.history.pop()
        self.col_heights[col] -= 1
        move = 1 << (self.col_heights[col])
        self.board_state[opp] ^= move
        self.moves -= 1

    def winning_board_state(self):
        ''' returns true if last played column creates winning alignment '''
        opp = self.get_opponent()
        for shift in self.bit_shifts:
            test = self.board_state[opp] & (self.board_state[opp] >> shift)
            if test & (test >> 2 * shift):
                return True
        return False

    def get_score(self):
        ''' returns score of complete game (evaluated for winning opponent) '''
        return - (self.w * self.h + 1 - self.moves) // 2

    def __get_bit_shifts(self):
        return [
            1,              # | vertical
            self.h,         # \ diagonal
            self.h + 1,     # - horizontal
            self.h + 2      # / diagonal
        ]

    def __get_base_search_order(self):
        base_search_order = list(range(self.w))
        base_search_order.sort(key=lambda x: abs(self.w // 2 - x))
        return base_search_order

    def __col_sort(self, col):
        player = self.get_current_player()
        move = 1 << self.col_heights[col]
        count = 0
        state = self.board_state[player] | move

        for shift in self.bit_shifts:
            test = state & (state >> shift) & (state >> 2 * shift)
            if test:
                count += bin(test).count('1')

        return count



import os
import sys

import numpy as np
from mu_alpha_zero.General.arena import GeneralArena

sys.path.append(os.path.abspath("."))
from typing import Type

import torch as th
from mu_alpha_zero.AlphaZero.Arena.players import NetPlayer
from mu_alpha_zero.trainer import Trainer
from mu_alpha_zero.General.utils import find_project_root, net_not_none
from mu_alpha_zero.General.mz_game import MuZeroGame
from mu_alpha_zero.General.memory import GeneralMemoryBuffer
from mu_alpha_zero.General.network import GeneralNetwork
from mu_alpha_zero.MuZero.MZ_Arena.arena import MzArena
from mu_alpha_zero.MuZero.MZ_MCTS.mz_search_tree import MuZeroSearchTree
from mu_alpha_zero.config import MuZeroConfig
from mu_alpha_zero.Hooks.hook_manager import HookManager


class MuZero:
    """
    Class for managing the training and creation of a MuZero model.

    Attributes:
        game_manager (MuZeroGame): The game manager instance.
        net (GeneralNetwork): The neural network used by MuZero for prediction.
        trainer (Trainer): The trainer object responsible for training the model.
        device (torch.device): The device (CPU or CUDA) used for computations.

    Methods:
        __init__(self, game_manager: MuZeroGame)
            Initializes a new instance of the MuZero class.

        create_new(self, args: dict, network_class: Type[GeneralNetwork], headless: bool = True,
                   checkpointer_verbose: bool = False)
            Creates a new MuZero model using the specified arguments.

        train(self)
            Trains the MuZero model.
    """

    def __init__(self, game_manager: MuZeroGame):
        self.game_manager = game_manager
        self.net = None
        self.trainer = None
        self.device = th.device("cuda" if th.cuda.is_available() else "cpu")
        self.tree = None
        self.muzero_config = None

    def create_new(self, muzero_config: MuZeroConfig, network_class: Type[GeneralNetwork], memory: GeneralMemoryBuffer,
                   hook_manager: HookManager or None = None,
                   headless: bool = True,
                   arena_override: GeneralArena or None = None,
                   checkpointer_verbose: bool = False):
        muzero_config.net_action_size = int(self.game_manager.get_num_actions())
        if not os.path.isabs(muzero_config.pickle_dir):
            muzero_config.pickle_dir = find_project_root() + "/" + muzero_config.pickle_dir
        self.muzero_config = muzero_config
        network = network_class.make_from_config(muzero_config, hook_manager=hook_manager).to(
            self.device)
        self.tree = MuZeroSearchTree(self.game_manager.make_fresh_instance(), muzero_config)

        net_player = NetPlayer(self.game_manager.make_fresh_instance(),
                               **{"network": network, "monte_carlo_tree_search": self.tree})

        arena = MzArena(self.game_manager.make_fresh_instance(), self.muzero_config,
                        self.device) if arena_override is None else arena_override
        self.trainer = Trainer.create(self.muzero_config, self.game_manager.make_fresh_instance(), network, self.tree,
                                      net_player,
                                      headless=headless,
                                      checkpointer_verbose=checkpointer_verbose, arena_override=arena,
                                      hook_manager=hook_manager,
                                      memory_override=memory)
        self.net = self.trainer.get_network()

    def from_checkpoint(self, network_class: Type[GeneralNetwork], memory: GeneralMemoryBuffer, path: str,
                        checkpoint_dir: str,
                        headless: bool = True,
                        hook_manager: HookManager or None = None,
                        logdir_override: str = None,
                        checkpointer_verbose: bool = False):
        self.trainer = Trainer.from_checkpoint(network_class, MuZeroSearchTree, NetPlayer, path,
                                               checkpoint_dir,
                                               self.game_manager,
                                               headless=headless,
                                               hook_manager=hook_manager,
                                               checkpointer_verbose=checkpointer_verbose,
                                               mem=memory, log_dir_override=logdir_override)
        self.net = self.trainer.get_network()
        self.tree = self.trainer.get_tree()
        self.args = self.trainer.get_args()

    def train(self):
        net_not_none(self.net)
        self.trainer.train()

    def train_parallel(self, use_reanalyze: bool):
        net_not_none(self.net)
        self.trainer.train_parallel(use_reanalyze=use_reanalyze, use_pitting=False)

    def predict(self, x: np.ndarray, tau: float = 0) -> int:
        net_not_none(self.net)
        assert x.shape == self.muzero_config.target_resolution + (
            3,), "Input shape must match target resolution with 3 channels. Got: " + str(x.shape)
        self.net.eval()
        pi, (v, _) = self.tree.search(self.net, x, None, self.device, tau=tau)
        move = self.game_manager.select_move(pi, tau=self.muzero_config.tau)
        return move



import copy
import multiprocess
# from pathos.helpers import mp as multiprocess
from multiprocess.managers import BaseManager
from mu_alpha_zero.mem_buffer import MemBuffer


class SharedStorage:
    def __init__(self, mem_buffer: MemBuffer):
        self.mem_buffer = mem_buffer
        self.experimental_network_params: dict or None = None
        self.stable_network_params: dict or None = None
        self.lock = multiprocess.context._default_context.Lock()
        self.optimizer_state_dict = None
        self.was_pitted = True
        self.training_iter = 0
        self.combined_losses = []

    def get_experimental_network_params(self):
        return copy.deepcopy(self.experimental_network_params)

    def set_experimental_network_params(self, network_params: dict or None):
        self.experimental_network_params = copy.deepcopy(network_params)

    def get_stable_network_params(self):
        return copy.deepcopy(self.stable_network_params)

    def set_stable_network_params(self, network_params: dict):
        self.stable_network_params = copy.deepcopy(network_params)

    def eval_length(self):
        with self.lock:
            return self.mem_buffer.eval_length()

    def train_length(self):
        with self.lock:
            return self.mem_buffer.train_length()

    def add_list(self, *args, **kwargs):
        with self.lock:
            return self.mem_buffer.add_list(*args, **kwargs)

    def get_dir_path(self):
        with self.lock:
            return self.mem_buffer.get_dir_path()

    def get_buffer(self):
        with self.lock:
            return self.mem_buffer.get_buffer()

    def set_optimizer(self, optimizer):
        self.optimizer_state_dict = copy.deepcopy(optimizer)

    def get_optimizer(self):
        return copy.deepcopy(self.optimizer_state_dict)

    def set_was_pitted(self, was_pitted: bool):
        self.was_pitted = was_pitted

    def get_was_pitted(self):
        return self.was_pitted

    def batch_with_priorities(self, *args, **kwargs):
        with self.lock:
            return self.mem_buffer.batch_with_priorities(*args, **kwargs)

    def reset_priorities(self):
        with self.lock:
            return self.mem_buffer.reset_priorities()

    def batch(self, *args, **kwargs):
        with self.lock:
            return self.mem_buffer.batch(*args, **kwargs)

    def set_training_iter(self, training_iter):
        with self.lock:
            self.training_iter = training_iter

    def get_training_iter(self):
        with self.lock:
            return self.training_iter

    def add_combined_loss(self, loss: float):
        with self.lock:
            self.combined_losses.append(loss)

    def get_combined_losses(self):
        with self.lock:
            return self.combined_losses


class SharedStorageManager(BaseManager):
    pass


SharedStorageManager.register("SharedStorage", SharedStorage)
SharedStorageManager.register("MemBuffer", MemBuffer)



import json
import os
import shutil
import time
from typing import Type, Literal, Callable

import numpy as np
import optuna
# import pygraphviz
import torch as th

from mu_alpha_zero.AlphaZero.constants import SAMPLE_AZ_ARGS as test_args
from mu_alpha_zero.mem_buffer import MemBuffer
from mu_alpha_zero.config import Config, AlphaZeroConfig
from mu_alpha_zero.General.network import GeneralNetwork


class DotDict(dict):
    def __getattr__(self, name):
        return self[name]

    def __setattr__(self, name, value):
        self[name] = value


def augment_experience_with_symmetries(game_experience: list, board_size) -> list:
    game_experience_ = []
    for state, pi, v, _ in game_experience:
        pi = np.array([x for x in pi.values()])
        game_experience_.append((state, pi, v))
        for axis, k in zip([0, 1], [1, 3]):
            state_ = np.rot90(state.copy(), k=k)
            pi_ = np.rot90(pi.copy().reshape(board_size, board_size), k=k).flatten()
            game_experience_.append((state_, pi_, v))
            del state_, pi_
            state_ = np.flip(state.copy(), axis=axis)
            pi_ = np.flip(pi.copy().reshape(board_size, board_size), axis=axis).flatten()
            game_experience_.append((state_, pi_, v))

    return game_experience_


def rotate_stack(state: np.ndarray, k: int):
    for dim in range(state.shape[0]):
        state[dim] = np.rot90(state[dim], k=k)
    return state


def flip_stack(state: np.ndarray, axis: int):
    for dim in range(state.shape[0]):
        state[dim] = np.flip(state[dim], axis=axis)
    return state


def make_channels(game_experience: list):
    experience = []
    for state, pi, v, current_player, move in game_experience:
        state = make_channels_from_single(state)
        experience.append((state, pi, v, current_player, move))

    return experience


def make_channels_from_single(state: np.ndarray):
    player_one_state = np.where(state == 1, 1, 0)  # fill with 1 where player 1 has a piece else 0
    player_minus_one_state = np.where(state == -1, 1, 0)  # fill with 1 where player -1 has a piece else 0
    empty_state = np.where(state == 0, 1, 0)  # fill with 1 where empty spaces else 0
    return np.stack([state, player_one_state, player_minus_one_state, empty_state], axis=0)


def mask_invalid_actions(probabilities: np.ndarray, mask: np.ndarray) -> np.ndarray:
    mask = mask.reshape(probabilities.shape)
    valids = probabilities * mask
    valids_sum = valids.sum()
    return valids / valids_sum
    # to_print = ""  # for debugging
    # # mask = np.where(observations != 0, -5, observations)
    # # mask = np.where(mask == 0, 1, mask)
    # # mask = np.where(mask == -5, 0, mask)
    # valids = probabilities.reshape(-1, board_size ** 2) * mask.reshape(-1, board_size ** 2)
    # valids_sum = valids.sum()
    # if valids_sum == 0:
    #     # When no valid moves are available (shouldn't happen) sum of valids is 0, making the returned valids an array
    #     # of nan's (result of division by zero). In this case, we create a uniform probability distribution.
    #     to_print += f"Sum of valid probabilities is 0. Creating a uniform probability...\nMask:\n{mask}"
    #     valids = np.full(valids.shape, 1.0 / np.prod(valids.shape))
    # else:
    #     valids = valids / valids_sum  # normalize
    #
    # if len(to_print) > 0:
    #     print(to_print, file=open("masking_message.txt", "w"))
    # return valids


def mask_invalid_actions_batch(states: th.tensor) -> th.tensor:
    masks = []
    for state in states:
        np_state = state.detach().cpu().numpy()
        mask = np.where(np_state != 0, -5, np_state)
        mask = np.where(mask == 0, 1, mask)
        mask = np.where(mask == -5, 0, mask)
        masks.append(mask)

    return th.tensor(np.array(masks), dtype=th.float32).squeeze(1)


def check_args(args: dict):
    required_keys = ["num_net_channels", "num_net_in_channels", "net_dropout", "net_action_size", "num_simulations",
                     "self_play_games", "num_iters", "epochs", "lr", "max_buffer_size", "num_pit_games",
                     "random_pit_freq", "board_size", "batch_size", "tau", "c", "checkpoint_dir", "update_threshold"]

    for key in required_keys:
        if key not in args:
            raise KeyError(f"Missing key {key} in args dict. Please supply all required keys.\n"
                           f"Required keys: {required_keys}.")


def calculate_board_win_positions(n: int, k: int):
    return get_num_horizontal_conv_slides(n, k) * (n * 2 + 2) + 4 * sum(
        [get_num_horizontal_conv_slides(x, k) for x in range(k, n)])


def get_num_horizontal_conv_slides(board_size: int, kernel_size: int) -> int:
    return (board_size - kernel_size) + 1


def az_optuna_parameter_search(n_trials: int, target_values: list, target_game, config: AlphaZeroConfig,
                               net_class: Type[GeneralNetwork], results_dir: str, az,refresh_az: Callable):
    """
    Performs a hyperparameter search using optuna. This method is meant to be called using the start_jobs.py script.
    For this method to work, a mysql database must be running on the storage address and an optuna study with the
    given name and the 'maximize' direction must exist.

    :param n_trials: num of trials to run the search for.
    :param config: The config to use for the search.
    :return:
    """

    def get_function_from_value(value, trial: optuna.Trial):
        if type(value[1]) == int:
            return trial.suggest_int(value[0], value[1], value[2])
        if type(value[1]) == float:
            return trial.suggest_float(value[0], value[1], value[2])
        if type(value[1]) == list:
            return trial.suggest_categorical(value[0], value[1])

    def objective(trial: optuna.Trial):
        az = refresh_az()
        for value in target_values:
            setattr(config, value[0], get_function_from_value(value, trial))

        az.trainer.opponent_network.load_state_dict(az.trainer.network.state_dict())
        shared_storage_manager = SharedStorageManager()
        shared_storage_manager.start()
        mem = shared_storage_manager.MemBuffer(az.trainer.memory.max_size, az.trainer.memory.disk,
                                               az.trainer.memory.full_disk,
                                               az.trainer.memory.dir_path, hook_manager=az.trainer.memory.hook_manager)
        shared_storage: SharedStorage = shared_storage_manager.SharedStorage(mem)
        shared_storage.set_stable_network_params(az.trainer.network.state_dict())
        pool = az.trainer.mcts.start_continuous_self_play(
            az.trainer.make_n_networks(az.trainer.muzero_alphazero_config.num_workers),
            az.trainer.make_n_trees(az.trainer.muzero_alphazero_config.num_workers),
            shared_storage, az.trainer.device,
            az.trainer.muzero_alphazero_config,
            az.trainer.muzero_alphazero_config.num_workers,
            az.trainer.muzero_alphazero_config.num_worker_iters)
        az.trainer.logger.log(
            f"Successfully started a pool of {az.trainer.muzero_alphazero_config.num_workers} workers for "
            f"self-play (1/2).")
        p2 = Process(target=az.trainer.network.continuous_weight_update,
                     args=(
                         shared_storage, az.trainer.muzero_alphazero_config, az.trainer.checkpointer,
                         az.trainer.logger))

        p2.start()
        p4 = Process(target=az.trainer.arena.continuous_pit, args=(
            az.trainer.net_player.make_fresh_instance(),
            az.trainer.net_player.make_fresh_instance(),
            RandomPlayer(az.trainer.game_manager.make_fresh_instance(), **{}),
            az.trainer.muzero_alphazero_config.num_pit_games,
            az.trainer.muzero_alphazero_config.num_simulations,
            shared_storage,
            az.trainer.checkpointer,
            False,
            1
        ))
        p4.start()
        last_len = 0
        max_len = 500
        while len(shared_storage.get_combined_losses()) < max_len:
            if len(shared_storage.get_combined_losses()) <= last_len:
                time.sleep(2)
                continue
            last_len = len(shared_storage.get_combined_losses())
            trial.report(shared_storage.get_combined_losses()[-1], len(shared_storage.get_combined_losses()))
        pool.terminate()
        p2.terminate()
        p4.terminate()
        return shared_storage.get_combined_losses()[-1]

    from mu_alpha_zero.shared_storage_manager import SharedStorageManager, SharedStorage
    from mu_alpha_zero.AlphaZero.Arena.players import RandomPlayer
    from mu_alpha_zero.mem_buffer import MemBuffer
    from multiprocess.context import Process
    config.show_tqdm = False
    study = optuna.create_study(study_name="AlphaZeroHyperparameterSearch", direction="minimize")
    study.optimize(objective, n_trials=n_trials)
    with open(f"{results_dir}/best_params.json", "w") as f:
        json.dump(study.best_params, f)


def build_net_from_config(muzero_config: Config, device):
    from mu_alpha_zero.AlphaZero.Network.nnet import AlphaZeroNet
    network = AlphaZeroNet(muzero_config.num_net_in_channels, muzero_config.num_net_channels,
                           muzero_config.net_dropout, muzero_config.net_action_size,
                           muzero_config.az_net_linear_input_size)
    return network.to(device)


def build_all_from_config(muzero_alphazero_config: Config, device, lr=None, buffer_size=None):
    if lr is None:
        lr = muzero_alphazero_config.lr
    if buffer_size is None:
        buffer_size = muzero_alphazero_config.max_buffer_size
    network = build_net_from_config(muzero_alphazero_config, device)
    optimizer = th.optim.Adam(network.parameters(), lr=lr)
    memory = MemBuffer(max_size=buffer_size)
    return network, optimizer, memory


def make_net_from_checkpoint(checkpoint_path: str, args: DotDict | None):
    if args is None:
        args = DotDict(test_args)
    device = th.device("cuda" if th.cuda.is_available() else "cpu")
    net = build_net_from_config(args, device)
    data = th.load(checkpoint_path)
    net.load_state_dict(data["net"])
    return net




def visualize_tree(root_node, output_file_name: str, depth_limit: int | None = None):
    graph = pygraphviz.AGraph()
    graph.graph_attr["label"] = "MCTS visualization"
    graph.node_attr["shape"] = "circle"
    graph.edge_attr["color"] = "blue"
    graph.node_attr["color"] = "gold"
    if depth_limit is None:
        depth_limit = float("inf")

    def make_graph(node, parent, g: pygraphviz.AGraph, d_limit: int):

        state_ = None
        if node.state is None:
            state_ = str(np.random.randint(low=0, high=5, size=parent.state.shape))
        else:
            state_ = str(node.state)
        g.add_node(state_)
        if parent != node:
            g.add_edge(str(parent.state), state_)
        if not node.was_visited() or d_limit <= 0:
            return
        # queue_ = deque(root_node.children.values())
        # depth = 1
        # num_children = 25
        # children_iterated = 0
        # parent = root_node

        for child in node.children.values():
            make_graph(child, node, g, d_limit=d_limit - 1 if depth_limit != float("inf") else depth_limit)

    make_graph(root_node, root_node, graph, d_limit=depth_limit)
    graph.layout(prog="dot")
    graph.draw(f"{output_file_name}.png")


def cpp_data_to_memory(data: list, memory: MemBuffer, board_size: int):
    # import pickle
    # test_data = pickle.load(open(f"{find_project_root()}/history.pkl","rb"))
    for game_data in data:
        for state, pi, v in game_data:
            state = th.tensor(state, dtype=th.float32).reshape(board_size, board_size)
            pi = th.tensor(pi, dtype=th.float32)
            memory.add((state, pi, v))

#
# if __name__ == "__main__":
#     cpp_data_to_memory(None,MemBuffer(max_size=10_000),test_args)



import gzip
import os
import uuid

import numpy as np


class LazyArray:
    def __init__(self, array: np.ndarray, directory_path: str):
        self.directory_path = directory_path
        self.path = f"{directory_path}/array_{uuid.uuid4()}.npy.gz"
        self.persistent = False
        self.save_array(array)

    def load_array(self):
        f = gzip.GzipFile(self.path, 'r')
        arr = np.load(f)
        f.close()
        return arr

    def save_array(self, array: np.ndarray):
        file = gzip.GzipFile(self.path, 'w')
        np.save(file, array)
        file.close()

    def __array__(self):
        return self.load_array()

    def remove_array(self):
        os.remove(self.path)

    def make_persistent(self):
        self.persistent = True





from abc import ABC, abstractmethod

import numpy as np
import torch as th
from mu_alpha_zero.General.utils import adjust_probabilities


class MuZeroGame(ABC):
    """
    Keep state representations as Gzip if, muzero_config.use_true_game_state is True
    """

    @abstractmethod
    def get_next_state(self, action: int, player: int or None) -> (
            np.ndarray or th.Tensor, int, bool):
        """
        Given a game state and an action return the next state. Currently this implementation only supports one player, so supply None.

        :param state: The current state of the game.
        :param action: The action to be taken.
        :param player: The player taking the action.
        :return: The next state, the reward and a boolean indicating if the game is done.
        """
        pass

    @abstractmethod
    def reset(self) -> np.ndarray or th.Tensor:
        """
        Resets the game and returns the initial board.
        """
        pass

    @abstractmethod
    def get_noop(self) -> int:
        """
        Returns the noop action.
        """
        pass

    @abstractmethod
    def get_num_actions(self) -> int:
        """
        Returns the number of actions possible in the current environment.
        """
        pass

    @abstractmethod
    def game_result(self, player: int or None) -> bool or None:
        """
        Returns true if this environment is done, false otherwise. If this environment doesn't have a clear terminal
        state, return None.  Currently this implementation only supports one player, so supply None.
        """
        pass

    @abstractmethod
    def make_fresh_instance(self):
        """
        Return fresh instance of this game manager.
        """
        pass

    @abstractmethod
    def render(self):
        """
        Render the GUI of the game.
        """
        pass

    @abstractmethod
    def frame_skip_step(self, action: int, player: int or None, frame_skip: int = 4) -> (
            np.ndarray or th.Tensor, int, bool):
        """
        Wrapper for the frame skip method. This method should be used instead of get_next_state.
        """
        pass

    @staticmethod
    def select_move(action_probs: dict, tau: float):
        """
        Samples a move from the action probabilities.
        """
        action_probs = adjust_probabilities(action_probs, tau=tau)
        moves, probs = zip(*action_probs.items())
        return np.random.choice(moves, p=probs)

    @abstractmethod
    def get_invalid_actions(self, player: int):
        """
        Returns a mask where invalid actions == 0 and valid == 1 as a numpy array over the action space. The state is managed internally.
        """
        pass

    @abstractmethod
    def get_random_valid_action(self, state: np.ndarray, **kwargs):
        """
        Returns a random valid action in the given state and optionally the player.
        """
        pass

    @abstractmethod
    def get_state_for_passive_player(self, state: np.ndarray, player: int):
        """
        Returns a state with information specific for the provided player. This should only be done when a part of
        the actual observation is player specific (for example the player's cards in a card game). In other cases
        this method should return the provided state as is (for example in chess or other board games).
        """
        pass



from abc import ABC, abstractmethod

from mu_alpha_zero.General.az_game import AlphaZeroGame

from mu_alpha_zero.General.mz_game import MuZeroGame

from mu_alpha_zero.Hooks.hook_manager import HookManager
from mu_alpha_zero.config import Config
import torch as th


class GeneralNetwork(ABC):
    @abstractmethod
    def make_fresh_instance(self):
        """
        Returns a fresh instance of the network
        """
        pass

    @classmethod
    @abstractmethod
    def make_from_config(cls, config: Config, hook_manager: HookManager or None = None):
        """
        Builds the network from the given arguments dict.
        """
        pass

    @abstractmethod
    def train_net(self, memory_buffer, muzero_alphazero_config: Config) -> tuple[float, list[float]]:
        """
        Trains the network for given number of epochs
        """
        pass

    @abstractmethod
    def eval_net(self, memory_buffer, muzero_alphazero_config: Config) -> None:
        """
        Evaluates the network against the evaluation dataset and reports directly to wandb.
        :param memory_buffer: The memory buffer where the datasets should be pulled from.
        :param muzero_alphazero_config: The config for this algorithm.
        :return: None, reports to wandb.
        """
        pass


class GeneralMuZeroNetwork(GeneralNetwork):

    @abstractmethod
    def dynamics_forward(self, x: th.Tensor, predict: bool = False) -> (th.Tensor, th.Tensor):
        """
        Forward pass for the dynamics network. Should operate only on the hidden state, not the actual game state.
        Call representation_forward first.
        :param x: The input hidden state.
        :param predict: Whether to predict or do pure forward pass.
        :return: A tuple of the next hidden state and the immediate reward.
        """
        pass

    @abstractmethod
    def prediction_forward(self, x: th.Tensor, predict: bool = False) -> (th.Tensor, th.Tensor):
        """
        Forward pass for the prediction network. As dynamics forward, this should operate only on the hidden state,
        not the actual game state.
        Call representation_forward first.
        :param x: The input hidden state.
        :param predict: Whether to predict or do pure forward pass.
        :return: A tuple of the action probability distribution and the value of the current state.
        """
        pass

    @abstractmethod
    def representation_forward(self, x: th.Tensor) -> th.Tensor:
        """
        Forward pass for the representation network. This should operate on the actual game state.
        :param x: The input game state.
        :return: The hidden state.
        """
        pass

    @abstractmethod
    def muzero_loss(self, y_hat: th.Tensor, y: th.Tensor) -> th.Tensor:
        """
        Calculate the loss for the action probability distribution.
        :param y_hat: The predicted action probability distribution.
        :param y: The MCTS improved action probability distribution.
        :return: The loss.
        """
        pass


class GeneralAlphZeroNetwork(GeneralNetwork):
    @abstractmethod
    def predict(self, x: th.Tensor, muzero: bool = True) -> (th.Tensor, th.Tensor):
        """
        Predict the action probability distribution and the value of the current state.
        :param x: The input game state.
        :param muzero: Whether to predict for MuZero or AlphaZero.
        :return: A tuple of the action probability distribution and the value of the current state.
        """
        pass



import copy
from typing import Tuple, List, Any, Dict

import wandb
from multiprocess import set_start_method

set_start_method("spawn", force=True)
from multiprocess.pool import Pool

import numpy as np
import torch as th

from mu_alpha_zero.AlphaZero.MCTS.az_node import AlphaZeroNode
from mu_alpha_zero.AlphaZero.utils import augment_experience_with_symmetries, mask_invalid_actions
from mu_alpha_zero.Game.tictactoe_game import TicTacToeGameManager
from mu_alpha_zero.General.memory import GeneralMemoryBuffer
from mu_alpha_zero.General.network import GeneralNetwork
from mu_alpha_zero.General.search_tree import SearchTree
from mu_alpha_zero.Hooks.hook_manager import HookManager
from mu_alpha_zero.Hooks.hook_point import HookAt
from mu_alpha_zero.config import AlphaZeroConfig
from mu_alpha_zero.mem_buffer import SingleGameData, DataPoint
from mu_alpha_zero.shared_storage_manager import SharedStorage


class McSearchTree(SearchTree):
    def __init__(self, game_manager: TicTacToeGameManager, alpha_zero_config: AlphaZeroConfig,
                 hook_manager: HookManager or None = None):
        self.game_manager = game_manager
        self.alpha_zero_config = alpha_zero_config
        self.hook_manager = hook_manager if hook_manager is not None else HookManager()
        self.root_node = None

    def play_one_game(self, network: GeneralNetwork, device: th.device) -> tuple[
        list | list[tuple[int | Any, Any, float, int]], int, int, int]:
        """
        Plays a single game using the Monte Carlo Tree Search algorithm.

        Args:
            network: The neural network used for searching and evaluating moves.
            device: The device (e.g., CPU or GPU) on which the network is located.

        Returns:
            A tuple containing the game history, and the number of wins, losses, and draws.
            The game history is a list of tuples, where each tuple contains:
            - The game state multiplied by the current player
            - The policy vector (probability distribution over moves)
            - The game result (1 for a win, -1 for a loss, 0 for a draw)
            - The current player (-1 or 1)
        """
        # tau = self.args["tau"]
        state = self.game_manager.reset()
        current_player = 1
        game_history = []
        # game_data = SingleGameData()
        results = {"1": 0, "-1": 0, "D": 0}
        move_number = 0
        tau = self.alpha_zero_config.tau
        while True:
            if move_number > self.alpha_zero_config.zero_tau_after_first_n_moves != 0:
                tau = 0
            pi, _ = self.search(network, state, current_player, device)
            move = self.game_manager.select_move(pi, tau=tau)
            # self.step_root([move])
            self.step_root(None)
            # pi = [x for x in pi.values()]
            game_history.append((state * current_player, pi, None, current_player,
                                 self.game_manager.get_invalid_actions(state, current_player)))
            state = self.game_manager.get_next_state(state, self.game_manager.network_to_board(move), current_player)
            r = self.game_manager.game_result(current_player, state)
            if r is not None:
                if r == current_player:
                    results["1"] += 1
                elif -1 < r < 1:
                    results["D"] += 1
                else:
                    results["-1"] += 1

                if -1 < r < 1:
                    game_history = [(x[0], x[1], r, x[3], x[4]) for x in game_history]
                else:
                    game_history = [(x[0], x[1], r * current_player * x[3], x[3], x[4]) for x in game_history]

                break
            current_player *= -1
            move_number += 1

        # game_history = make_channels(game_history)
        if self.alpha_zero_config.augment_with_symmetries:
            game_history = self.game_manager.augment_game_history_with_symmetries(game_history)
        self.hook_manager.process_hook_executes(self, self.play_one_game.__name__, __file__, HookAt.TAIL,
                                                args=(game_history, results))
        # for state, pi, r, player, move_mask in game_history:
        #     game_data.add_data_point(DataPoint(pi, r, None, None, player, state, move_mask))
        return game_history, results["1"], results["-1"], results["D"]

    def search(self, network, state, current_player, device, tau=None):
        """
        Perform a Monte Carlo Tree Search on the current state starting with the current player.
        :param tau:
        :param network:
        :param state:
        :param current_player:
        :param device:
        :return:
        """
        num_simulations = self.alpha_zero_config.num_simulations
        if tau is None:
            tau = self.alpha_zero_config.tau
        if self.root_node is None:
            self.root_node = AlphaZeroNode(current_player, times_visited_init=0)
        state_ = self.game_manager.get_canonical_form(state, current_player)
        # state_ = make_channels_from_single(state_)
        state_ = th.tensor(state_, dtype=th.float32, device=device).unsqueeze(0)
        probabilities, v = network.predict(state_, muzero=False)
        if self.alpha_zero_config.add_dirichlet_noise:
            probabilities = (
                                    1 - self.alpha_zero_config.dirichlet_alpha) * probabilities + self.alpha_zero_config.dirichlet_alpha * np.random.dirichlet(
                [0.03] * len(probabilities))
        probabilities = mask_invalid_actions(probabilities,
                                             self.game_manager.get_invalid_actions(state.copy(), current_player))
        probabilities = probabilities.flatten().tolist()
        self.root_node.expand(state, probabilities)
        for simulation in range(num_simulations):
            current_node = self.root_node
            path = [current_node]
            action = None
            while current_node.was_visited():
                current_node, action = current_node.get_best_child(c=self.alpha_zero_config.c)
                if current_node is None:  # This was for testing purposes
                    th.save(self.root_node, "root_node.pt")
                    th.save(network.state_dict(), f"network_none_checkpoint_{current_player}.pt")
                    raise ValueError("current_node is None")
                path.append(current_node)

            # leaf node reached
            next_state = self.game_manager.get_next_state(current_node.parent().state.copy(),
                                                          self.game_manager.network_to_board(action),
                                                          current_node.parent().current_player)
            next_state_ = self.game_manager.get_canonical_form(next_state, current_node.current_player)
            v = self.game_manager.game_result(current_node.current_player, next_state)
            if v is None:
                # next_state_ = make_channels_from_single(next_state_)
                next_state_ = th.tensor(next_state_, dtype=th.float32, device=device).unsqueeze(0)
                probabilities, v = network.predict(next_state_, muzero=False)
                probabilities = mask_invalid_actions(probabilities, self.game_manager.get_invalid_actions(next_state,
                                                                                                          current_node.current_player))
                v = v.flatten().tolist()[0]
                probabilities = probabilities.flatten().tolist()
                current_node.expand(next_state, probabilities)

            self.backprop(v, path)

        self.hook_manager.process_hook_executes(self, self.search.__name__, __file__, HookAt.TAIL,
                                                args=(tau, self.root_node))
        return self.root_node.get_self_action_probabilities(), None

    def backprop(self, v, path):
        """
        Backpropagates the value `v` through the search tree, updating the relevant nodes.

        Args:
            v (float): The value to be backpropagated.
            path (list): The path from the leaf node to the root node.

        Returns:
            None
        """
        for node in reversed(path):
            v *= -1
            node.total_value += v
            node.update_q(v)
            node.times_visited += 1

    def step_root(self, actions: list | None) -> None:
        if actions is not None:
            if self.root_node is not None:
                if not self.root_node.was_visited():
                    return
                for action in actions:
                    self.root_node = self.root_node.children[action]
                self.root_node.parent = None
        else:
            # reset root node
            self.root_node = None

    def make_fresh_instance(self):
        return McSearchTree(self.game_manager.make_fresh_instance(), self.alpha_zero_config)

    def self_play(self, net: GeneralNetwork, device: th.device, num_games: int, memory: GeneralMemoryBuffer) -> tuple[
        int, int, int]:
        wins_p1, wins_p2, draws = 0, 0, 0
        for game in range(num_games):
            game_results, wins_p1_, wins_p2_, draws_ = self.play_one_game(net, device)
            wins_p1 += wins_p1_
            wins_p2 += wins_p2_
            draws += draws_
            memory.add_list(game_results)

        return wins_p1, wins_p2, draws

    @staticmethod
    def parallel_self_play(nets: list, trees: list, memory: GeneralMemoryBuffer, device: th.device, num_games: int,
                           num_jobs: int):
        with Pool(num_jobs) as p:
            if not memory.is_disk:
                results = p.starmap(p_self_play,
                                    [(nets[i], trees[i], copy.deepcopy(device), num_games // num_jobs, None) for i in
                                     range(len(nets))])
            else:
                results = p.starmap(p_self_play, [
                    (nets[i], trees[i], copy.deepcopy(device), num_games // num_jobs, copy.deepcopy(memory)) for i in
                    range(len(nets))])
        wins_p1, wins_p2, draws = 0, 0, 0
        for result in results:
            wins_p1 += result[0]
            wins_p2 += result[1]
            draws += result[2]
            if not memory.is_disk:
                memory.add_list(result[3])
        return wins_p1, wins_p2, draws

    @staticmethod
    def start_continuous_self_play(nets: list, trees: list, shared_storage: SharedStorage, device: th.device,
                                   config: AlphaZeroConfig, num_jobs: int, num_worker_iters: int) -> Pool:
        pool = Pool(num_jobs)
        for i in range(num_jobs):
            pool.apply_async(c_p_self_play, args=(
                nets[i], trees[i], copy.deepcopy(device), config, i, shared_storage, num_worker_iters // num_jobs
            ))

        return pool


def p_self_play(net, tree, device, num_games, memory):
    wins_p1, wins_p2, draws = 0, 0, 0
    data = []
    for game in range(num_games):
        game_results, wp1, wp2, ds = tree.play_one_game(net, device)
        wins_p1 += wp1
        wins_p2 += wp2
        draws += ds
        if memory is not None:
            memory.add_list(game_results)
        else:
            data.extend(game_results)
    if memory is None:
        return wins_p1, wins_p2, draws, data
    return wins_p1, wins_p2, draws


def c_p_self_play(net, tree, device, config: AlphaZeroConfig, p_num: int, shared_storage: SharedStorage,
                  num_worker_iters: int):
    if p_num == 0:
        wandb.init(project=config.wandbd_project_name, name="Self play")

    net = net.to(device)
    for iter_ in range(num_worker_iters):
        if shared_storage.get_experimental_network_params() is None:
            params = shared_storage.get_stable_network_params()
        else:
            params = shared_storage.get_experimental_network_params()
        # params = shared_storage.get_stable_network_params()
        net.load_state_dict(params)
        net.eval()
        game_results, wins_p1, wins_p2, draws = tree.play_one_game(net, device)
        shared_storage.add_list(game_results)
        if p_num == 0:
            wandb.log({"Iteration": iter_, "Wins p1": wins_p1, "Wins p2": wins_p2, "Draws": draws})





board_size = 8

SAMPLE_AZ_ARGS = {
    "num_net_channels": 512,
    "num_net_in_channels": 1,
    "net_dropout": 0.3,
    "net_action_size": board_size ** 2,
    "num_simulations": 1317,
    "self_play_games": 300,
    "num_iters": 50,
    "epochs": 500,
    "lr": 0.0032485504583772953,
    "max_buffer_size": 100_000,
    "num_pit_games": 40,
    "random_pit_freq": 3,
    "board_size": board_size,
    "batch_size": 256,
    "tau": 1,
    "arena_tau": 0.04139160592420218,
    "c": 1,
    "checkpoint_dir": None,
    "update_threshold": 0.6,
    "minimax_depth": 4,
    "show_tqdm": True,
    "num_workers": 5,
    "num_to_win": 5,
    "log_epsilon": 1e-9,
    "zero_tau_after": 5,
    "az_net_linear_input_size": 18432,
    "log_dir": "Logs",
    "pushbullet_token": None
}

TRAINED_AZ_NET_ARGS = {
    "num_net_channels": 512,
    "num_net_in_channels": 1,
    "net_dropout": 0.3,
    "net_action_size": board_size ** 2,
    "num_simulations": 1317,
    "self_play_games": 3,
    "num_iters": 50,
    "epochs": 320,
    "lr": 0.0032485504583772953,
    "max_buffer_size": 100_000,
    "num_pit_games": 40,
    "random_pit_freq": 3,
    "board_size": board_size,
    "batch_size": 128,
    "tau": 1.0,
    "arena_tau": 0,  # 0.04139160592420218
    "c": 1.15,
    "checkpoint_dir": None,
    "update_threshold": 0.6,
    "minimax_depth": 4,
    "show_tqdm": True,
    "num_workers": 5,
    "num_to_win": 4,
    "log_epsilon": 1.4165210108199043e-08,
    "log_dir": "Logs",
    "pushbullet_token": None
}

SAMPLE_MZ_ARGS = {
    "num_net_channels": 512,
    "num_net_out_channels": 256,
    "num_net_in_channels": 1,
    "net_dropout": 0.3,
    "net_action_size": 14,
    "net_latent_size": 36,
    "num_simulations": 240,
    "self_play_games": 5,
    "K": 5,
    "gamma": 0.997,
    "frame_buffer_size": 32,
    "frame_skip": 4,
    "num_steps": 400,
    "num_iters": 50,
    "epochs": 100,
    "lr": 0.001,
    "max_buffer_size": 70_000,
    "num_pit_games": 40,
    "random_pit_freq": 2,
    "board_size": board_size,
    "batch_size": 255,
    "tau": 1,
    "arena_tau": 1e-2,
    "c": 1,
    "c2": 19652,
    "alpha": 0.8,
    "checkpoint_dir": None,
    "update_threshold": 0.6,
    "minimax_depth": None,  # don't use with muzero
    "show_tqdm": False,
    "num_workers": 5,
    "num_to_win": 5,
    "log_epsilon": 1e-9,
    "zero_tau_after": 5,
    "beta": 1,
    "env_id": "ALE/Asteroids-v5",
    "pickle_dir": "Pickles/Data",
    "target_resolution": (96, 96),
    "az_net_linear_input_size": 8192,
    "log_dir": "Logs",
    "pushbullet_token": None
}



from collections import OrderedDict


class LRUCache(OrderedDict):
    # limit size, removing the least recently used key when full
    # source: https://docs.python.org/3/library/collections

    def __init__(self, maxsize=128, *args, **kwds):
        self.maxsize = maxsize
        super().__init__(*args, **kwds)

    def __getitem__(self, key):
        value = super().__getitem__(key)
        self.move_to_end(key)
        return value

    def __setitem__(self, key, value):
        if key in self:
            self.move_to_end(key)
        super().__setitem__(key, value)
        if len(self) > self.maxsize:
            oldest = next(iter(self))
            del self[oldest]


def get_tt_entry(value, UB=False, LB=False):
    return {'value': value, 'UB': UB, 'LB': LB}


def solve(board):
    TT = LRUCache(4096)

    def recurse(alpha, beta):
        alpha_original = alpha

        # transposition table lookup
        if board.get_key() in TT:
            entry = TT[board.get_key()]
            if entry['LB']:
                alpha = max(alpha, entry['value'])  # lower bound stored in TT
            elif entry['UB']:
                beta = min(beta, entry['value'])    # upper bound stored in TT
            else:
                return entry['value']               # exact value stored in TT
            if alpha >= beta:
                return entry['value']               # cut-off (from TT)

        # negamax implementation
        if board.winning_board_state():
            return board.get_score()        # base case 1: winning alignment
        elif board.moves == board.w * board.h:
            return 0                        # base case 2: draw game
        value = -board.w * board.h
        for col in board.get_search_order():
            board.play(col)
            value = max(value, -recurse(-beta, -alpha))
            board.backtrack()
            alpha = max(alpha, value)
            if alpha >= beta:
                break                   # alpha cut-off

        # transposition table storage
        if value <= alpha_original:
            TT[board.get_key()] = get_tt_entry(value, UB=True)
        elif value >= beta:
            TT[board.get_key()] = get_tt_entry(value, LB=True)
        else:
            TT[board.get_key()] = get_tt_entry(value)       # store exact in TT

        return value
    return recurse(-1e9, 1e9)



import random
from typing import Type

import torch as th

from mu_alpha_zero.AlphaZero.Arena.players import Player, NetPlayer, RandomPlayer
from mu_alpha_zero.General.arena import GeneralArena
from mu_alpha_zero.General.mz_game import MuZeroGame
from mu_alpha_zero.Hooks.hook_manager import HookManager
from mu_alpha_zero.Hooks.hook_point import HookAt
from mu_alpha_zero.MuZero.utils import resize_obs, scale_state, scale_action
from mu_alpha_zero.config import MuZeroConfig
import wandb

from mu_alpha_zero.shared_storage_manager import SharedStorage


class MzArena(GeneralArena):
    def __init__(self, game_manager: MuZeroGame, muzero_config: MuZeroConfig, device: th.device,
                 hook_manager: HookManager or None = None):
        self.game_manager = game_manager
        self.muzero_config = muzero_config
        self.hook_manager = hook_manager if hook_manager is not None else HookManager()
        self.device = device

    def pit(self, player1: Type[Player], player2: Type[Player], num_games_to_play: int, num_mc_simulations: int,
            one_player: bool = False, start_player: int = 1):
        tau = self.muzero_config.arena_tau
        rewards = {1: [], -1: []}
        if one_player:
            num_games_per_player = num_games_to_play
        else:
            num_games_per_player = num_games_to_play // 2
        noop_num = random.randint(0, 30)
        players = {"1": player1, "-1": player2}
        for player in [1, -1]:
            kwargs = {"num_simulations": num_mc_simulations, "current_player": player, "device": self.device,
                      "tau": tau, "unravel": False}
            for game in range(num_games_per_player):
                self.game_manager.reset()

                state, _, _ = self.game_manager.frame_skip_step(self.game_manager.get_noop(), player,
                                                                frame_skip=noop_num)
                state = resize_obs(state, self.muzero_config.target_resolution, self.muzero_config.resize_images)
                state = scale_state(state, self.muzero_config.scale_state)
                for step in range(self.muzero_config.num_steps):
                    self.game_manager.render()
                    move = players[str(player)].choose_move(state, **kwargs)

                    state, reward, done = self.game_manager.frame_skip_step(move, player)
                    state = resize_obs(state, self.muzero_config.target_resolution, self.muzero_config.resize_images)
                    state = scale_state(state, self.muzero_config.scale_state)
                    try:
                        players[str(player)].monte_carlo_tree_search.buffer.add_frame(state, scale_action(move,
                                                                                                          self.game_manager.get_num_actions()),
                                                                                      player)
                    except AttributeError:
                        # Player is probably not net player and doesn't have monte_carlo_tree_search.
                        pass
                    rewards[player].append(reward)
                    if done:
                        break

        self.hook_manager.process_hook_executes(self, self.pit.__name__, __file__, HookAt.TAIL, (rewards, kwargs))
        return sum(rewards[1]), sum(rewards[-1]), 0

    def run_on_training_end(self):
        self.hook_manager.process_hook_executes(self, self.run_on_training_end.__name__, __file__, HookAt.ALL)





#######################        BOARD CLASS        ###########################
# The Board class is the data structure that holds the Connect 4 boards and the game operations
import numpy as np


# The Connect 4 board is 7 cells wide and 6 cells tall

# The underlying data structure is a 2-d list
# The first dimension is the column; the second dimension is the row

# Every cell in the above list contains either a 0 or a 1. Player 1 is represented by 0 tiles, and Player

##############################################################################
class Board:

    #static class variables - shared across all instances
    HEIGHT = 6
    WIDTH = 7

    def __init__(self, orig=None, hash=None):

        # copy
        if(orig):
            self.board = [list(col) for col in orig.board]
            self.numMoves = orig.numMoves
            self.lastMove = orig.lastMove
            return

        # creates from hash - NOTE: Does not understand move order
        elif(hash):
            self.board = []
            self.numMoves = 0
            self.lastMove = None

            # convert to base 3
            digits = []
            while hash:
                digits.append(int(hash % 3))
                hash //= 3

            col = []
            
            for item in digits:
                
                # 2 indicates new column
                if item == 2:
                    self.board.append(col)
                    col = []
                
                # otherwise directly append base number
                else:
                    col.append(item)
                    self.numMoves += 1
            return

        # create new
        else:
            self.board = [[] for x in range(self.WIDTH)]
            self.numMoves = 0
            self.lastMove = None
            return

    @classmethod
    def from_game_state(cls, game_state:np.ndarray, last_move:tuple):
        clss = cls()
        clss.board = [[int(x) if x == 1 else 0 for x in reversed(col.tolist()) if x != 0] for col in game_state.T]
        clss.numMoves = np.sum(game_state != 0)
        clss.lastMove = last_move
        return clss



    ########################################################################
    #                           Mutations
    ########################################################################

    # Puts a piece in the appropriate column and checks to see if it was a winning move
    # Pieces are either 1 or 0; automatically decided
    def makeMove(self, column):
        # update board data
        piece = self.numMoves % 2
        self.lastMove = (piece, column)
        self.numMoves += 1
        self.board[column].append(piece)


    ########################################################################
    #                           Observations
    ########################################################################

    # Generates a list of the valid children of the board
    # A child is of the form (move_to_make_child, child_object)
    def children(self):
        children = []
        for i in range(7):
            if len(self.board[i]) < 6:
                child = Board(self)
                child.makeMove(i)
                children.append((i, child))
        return children

    # Returns:
    #  -1 if game is not over
    #   0 if the game is a draw
    #   1 if player 1 wins
    #   2 if player 2 wins
    def isTerminal(self):
        for i in range(0,self.WIDTH):
            for j in range(0,self.HEIGHT):
                try:
                    if self.board[i][j]  == self.board[i+1][j] == self.board[i+2][j] == self.board[i+3][j]:
                        return self.board[i][j] + 1
                except IndexError:
                    pass

                try:
                    if self.board[i][j]  == self.board[i][j+1] == self.board[i][j+2] == self.board[i][j+3]:
                        return self.board[i][j] + 1
                except IndexError:
                    pass

                try:
                    if not j + 3 > self.HEIGHT and self.board[i][j] == self.board[i+1][j + 1] == self.board[i+2][j + 2] == self.board[i+3][j + 3]:
                        return self.board[i][j] + 1
                except IndexError:
                    pass

                try:
                    if not j - 3 < 0 and self.board[i][j] == self.board[i+1][j - 1] == self.board[i+2][j - 2] == self.board[i+3][j - 3]:
                        return self.board[i][j] + 1
                except IndexError:
                    pass
        if self.isFull():
            return 0
        return -1

    # Returns a unique decimal number for each board object based on the
    # underlying data
    def hash(self):

        power = 0
        hash = 0

        for column in self.board:

            # add 0 or 1 depending on piece
            for piece in column:
                hash += piece * (3 ** power)
                power += 1

            # add a 2 to indicate end of column
            hash += 2 * (3 ** power)
            power += 1

        return hash

    ########################################################################
    #                           Utilities
    ########################################################################

    # Return true iff the game is full
    def isFull(self):
        return self.numMoves == 42

    # Prints out a visual representation of the board
    # X's are 1's and 0's are 0s
    def print(self):
        print("")
        print("+" + "---+" * self.WIDTH)
        for rowNum in range(self.HEIGHT - 1, -1, -1):
            row = "|"
            for colNum in range(self.WIDTH):
                if len(self.board[colNum]) > rowNum:
                    row += " " + ('X' if self.board[colNum][rowNum] else 'O') + " |"
                else:
                    row += "   |"
            print(row)
            print("+" + "---+" * self.WIDTH)
        print(self.lastMove[1])
        print(self.numMoves)







import numpy as np
import torch as th
from gymnasium import make

from mu_alpha_zero.General.mz_game import MuZeroGame


class Asteroids(MuZeroGame):

    def get_state_for_passive_player(self, state: np.ndarray, player: int):
        return state

    def __init__(self):
        self.env = make("ALE/Asteroids-v5")
        self.done = False

    def reset(self) -> np.ndarray or th.Tensor:
        obs, _ = self.env.reset()
        return obs

    def get_noop(self) -> int:
        return 0

    def get_num_actions(self) -> int:
        return self.env.action_space.n

    def game_result(self, player: int or None) -> bool or None:
        return self.done

    def make_fresh_instance(self):
        return Asteroids()

    def get_next_state(self, action: int, player: int or None, frame_skip: int = 4) -> (
            np.ndarray or th.Tensor, int, bool):
        obs, rew, done, _, _ = self.env.step(action)
        self.done = done
        return obs, rew, done

    def frame_skip_step(self, action: int, player: int or None, frame_skip: int = 4) -> (
            np.ndarray or th.Tensor, int, bool):
        obs, rew, done = self.get_next_state(action, player)
        for i in range(frame_skip - 1):
            obs, rew, done = self.get_next_state(action, player)
        return obs, rew, done

    def render(self):
        pass

    def get_random_valid_action(self, state: np.ndarray, **kwargs):
        return self.env.action_space.sample()

    def get_invalid_actions(self, state: np.ndarray, player: int):
        return np.ones((self.get_num_actions()))



import math
import weakref
import numpy as np

from mu_alpha_zero.Game.tictactoe_game import TicTacToeGameManager as GameManager


class AlphaZeroNode:
    """
    This class defines a node in the search tree. It stores all the information needed for DeepMind's AlphaZero algorithm.
    """

    def __init__(self, current_player, select_probability=0, parent=None, times_visited_init=0):
        self.times_visited = times_visited_init
        self.was_init_with_zero_visits = times_visited_init == 0
        self.children = {}
        self.parent = weakref.ref(parent) if parent is not None else None
        self.select_probability = select_probability
        self.q = None
        self.current_player = current_player
        self.state = None
        self.total_value = 0

    def expand(self, state, action_probabilities) -> None:
        """
        Expands the newly visited node with the given action probabilities and state.
        :param state: np.ndarray of shape (board_size, board_size) representing the state current game board.
        :param action_probabilities: list of action probabilities for each action.
        :return: None
        """

        self.state = state.copy()  # copying here is probably not necessary.

        for action, probability in enumerate(action_probabilities):
            node = AlphaZeroNode(self.current_player * (-1), select_probability=probability, parent=self)
            self.children[action] = node

    def was_visited(self):
        return len(self.children) > 0

    def update_q(self, v):
        # Based on DeepMind's AlphaZero paper.
        if self.q is None:
            self.q = v
        else:
            self.q = (self.times_visited * self.q + v) / (self.times_visited + 1)

    def get_best_child(self, c=1.5):
        best_utc = -float("inf")
        best_child = None
        best_action = None
        valids_for_state = np.where(self.state != 0, -5, self.state)
        valids_for_state = np.where(valids_for_state == 0, 1, valids_for_state)
        utcs = []
        for action, child in self.children.items():
            action_ = np.unravel_index(action, self.state.shape)
            if valids_for_state[action_] != 1:
                continue
            child_utc = child.calculate_utc(c=c)
            utcs.append(child_utc)
            if child_utc > best_utc:
                best_utc = child_utc
                best_child = child
                best_action = action

        printable_children = [[child.state, child.times_visited, child.q, child.select_probability] for child in
                              self.children.values()]
        if best_child is None:
            # This was for testing purposes
            print("Best child is None. Possibly important info:\n", self.state, "\n",
                  valids_for_state, printable_children, self.was_visited(), utcs,
                  file=open("important_info"
                            ".txt", "w"))
            # raise Exception("Best child is None,terminating.")
        return best_child, best_action

    def calculate_utc(self, c=1.5):
        parent = self.parent()
        if self.q is None:
            # Inspiration taken from https://github.com/suragnair/alpha-zero-general/blob/master/MCTS.py
            utc = c * self.select_probability * math.sqrt(parent.times_visited + 1e-8)
        else:
            utc = self.q + c * (
                    self.select_probability * ((math.sqrt(parent.times_visited)) / (1 + self.times_visited)))

        return utc

    def get_self_value(self):
        return self.total_value / self.times_visited if self.times_visited > 0 else 0

    def get_self_action_probabilities(self):

        total_times_visited = self.times_visited
        action_probs = {}
        for action, child in self.children.items():
            action_probs[action] = child.times_visited / total_times_visited

        return action_probs

    def get_latent(self):
        return self.state
